from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from e3.anod.spec import Anod

if TYPE_CHECKING:
    from typing import Any, Dict, Final, List, Literal, Optional, Tuple, Union
    from e3.anod.package import Source, SourceBuilder
    from e3.collection.dag import DAG

    Choice = Union[Literal[0], Literal[1], Literal[2]]


class Action:
    """Action object.

    Action objects are used as node in DAG produced by anod scheduler. Only
    child classes are used directly.
    """

    __slots__ = ("uid", "data")

    def __init__(self, uid: str, data: Any):
        """Initialize an Action.

        :param uid: an unique identifier
        :param data: data associated with the node
        """
        self.uid = uid
        self.data = data

    @property
    def run_method(self) -> str:
        return "do_%s" % self.__class__.__name__.lower()


class Root(Action):
    """Root action.

    This action does not correspond to any real activity. It's only used as
    the root of the DAG and can be "executed" only once all the scheduled
    Actions are completed.
    """

    __slots__ = ("uid", "data")

    def __init__(self) -> None:
        """Initialize a root node."""
        super().__init__(uid="root", data="root")

    def __str__(self) -> str:
        return "root node"


class GetSource(Action):
    """GetSource Action.

    This action means that we need to retrieve a given source package. In the
    anod DAG context it can either have one child node: DownloadSource (read
    access to a store) or provide the choice between DownloadSource or
    CreateSource using a Decision node.
    """

    __slots__ = ("uid", "builder")

    def __init__(self, builder: SourceBuilder):
        """Object initializer.

        :param builder: A SourceBuilder object for the source we need to get.
        """
        super().__init__(uid="source_get.%s" % builder.name, data=builder)
        self.builder = builder

    def __str__(self) -> str:
        return "get source %s" % self.data.name


class Download(Action):
    """General root class for all download actions."""

    pass


class DownloadSource(Download):
    """DownloadSource Action.

    This action means the we need to download from the store a given
    source. DownloadSource is always a leaf of the DAG.
    """

    __slots__ = ("uid", "builder")

    def __init__(self, builder: SourceBuilder):
        """Object initializer.

        :param builder: A SourceBuilder object for the source we need
            to download.
        """
        super().__init__(uid="download.%s" % builder.name, data=builder)
        self.builder = builder

    def __str__(self) -> str:
        return "download source %s" % self.builder.name


class InstallSource(Action):
    """InstallSource Action.

    This means that a source should be installed. Child action is always
    a GetSource Action.
    """

    __slots__ = ("uid", "spec", "source")

    def __init__(self, uid: str, spec: Anod, source: Source):
        """Object initializer.

        :param uid: The job ID for this source's install.
        :param spec: The Anod instance of the spec providing those sources.
        :param source: The source we want to install.
        """
        super().__init__(uid, data=(spec, source))
        self.spec = spec
        self.source = source

    def __str__(self) -> str:
        return "install source %s" % self.data[1].name


class CreateSource(Action):
    """CreateSource Action.

    This means that we need to assemble the source package from a repositories
    checkouts. CreateSource has at least one Checkout child node.
    """

    __slots__ = ("uid", "anod_instance", "source_name")

    def __init__(self, anod_instance: Anod, source_name: str):
        """Initialize CreateSource object.

        :param anod_instance: The Anod instance of the spec providing
            the given source.
        :param source_name: name of source package to assemble
        """
        super().__init__(
            uid=f"{anod_instance.uid}.{source_name}", data=(anod_instance, source_name),
        )
        self.anod_instance = anod_instance
        self.source_name = source_name

    def __str__(self) -> str:
        return "create source %s" % self.data[1]


class CreateSources(Action):
    """CreateSources Action.

    This action does not correspond to any real activity. It's only
    used to group all CreateSource action corresponding to the same
    anod spec.
    """

    __slots__ = ("uid", "anod_instance")

    def __init__(self, anod_instance: Anod):
        """Initialize CreateSources object.

        :param anod_instance: the Anod instance of the spec
        """
        super().__init__(uid="%s.sources" % anod_instance.uid, data=(anod_instance))
        self.anod_instance = anod_instance

    def __str__(self) -> str:
        return "create all sources for %s.anod" % self.anod_instance.name


class Checkout(Action):
    """Checkout Action.

    This means that we need to perform the checkout/update of a given
    repository.
    """

    __slots__ = ("uid", "repo_name", "repo_data")

    def __init__(self, repo_name: str, repo_data: Dict[str, str]):
        """Initialize a Checkout object.

        :param repo_name: The name of the repository.
        :param repo_data: A dictionary with the following keys:
            - 'url': The repository URL;
            - 'revision': The revision to checkout;
            - 'vcs': The Version Control System kind (a string).
                At present, only 'git' is supported.
        """
        super().__init__(uid="checkout.%s" % repo_name, data=(repo_name, repo_data))
        self.repo_name = repo_name
        self.repo_data = repo_data

    def __str__(self) -> str:
        return "checkout %s" % self.data[0]


class AnodAction(Action):
    """AnodAction Action.

    Correspond to an Anod primitive call. Only subclasses should be used.
    """

    __slots__ = ("uid", "anod_instance")

    def __init__(self, anod_instance: Anod):
        """Initialize an anod Action.

        :param anod_instance: an Anod spec instance
        """
        assert isinstance(anod_instance, Anod)
        super().__init__(uid=anod_instance.uid, data=anod_instance)
        self.anod_instance = anod_instance

    def __str__(self) -> str:
        result = "{} {} for {}".format(
            self.data.kind, self.data.name, self.data.env.platform,
        )
        if self.data.qualifier:
            result += " (qualifier=%s)" % self.data.qualifier
        return result


class Build(AnodAction):
    """Anod build primitive."""

    pass


class Test(AnodAction):
    """Anod test primitive."""

    pass


class Install(AnodAction):
    """Anod install primitive."""

    pass


class DownloadBinary(Download):
    """DownloadBinary Action.

    Download a binary package from the store.
    """

    __slots__ = ("uid", "data")

    def __init__(self, data: Anod):
        """Initialize a DownloadBinary object.

        :param data: Anod instance
        """
        data_uid = data.uid.split(".")
        data_uid[-1] = "download_bin"
        uid = ".".join(data_uid)
        super().__init__(uid=uid, data=data)

    def __str__(self) -> str:
        return "download binary of %s" % self.uid.split(".", 1)[1].rsplit(".", 1)[0]


class Upload(Action):
    """General root class for all upload actions."""

    pass


class UploadComponent(Upload):
    """UploadComponent Action.

    Upload a component to the store.
    """

    __slots__ = ("uid", "data", "anod_instance")
    str_prefix = ""

    def __init__(self, data: Anod):
        """Initialize an UploadComponent object.

        :param data: Anod instance
        """
        data_uid = data.uid.split(".")
        data_uid[-1] = "upload_bin"
        uid = ".".join(data_uid)
        super().__init__(uid=uid, data=data)
        self.anod_instance = data

    def __str__(self) -> str:
        return "upload {} of {}".format(
            self.str_prefix, self.uid.split(".", 1)[1].rsplit(".", 1)[0],
        )


class UploadBinaryComponent(UploadComponent):
    """Upload binary component."""

    str_prefix = "binary package"


class UploadSourceComponent(UploadComponent):
    """Upload source only component."""

    str_prefix = "source metadata"


class UploadSource(Upload):
    """Upload a source package."""

    __slots__ = ("uid", "anod_instance", "source_name")

    def __init__(self, anod_instance: Anod, source_name: str):
        """Initialize UploadSource object.

        :param anod_instance: The Anod instance of the spec providing
            the given source.
        :param source_name: name of source package to assemble
        """
        instance_uid = anod_instance.uid.split(".")
        instance_uid[-1] = "upload_src"
        instance_uid.append(source_name)
        uid = ".".join(instance_uid)
        super().__init__(uid=uid, data=(anod_instance, source_name))
        self.anod_instance = anod_instance
        self.source_name = source_name

    def __str__(self) -> str:
        """Return string representation."""
        return "upload source %s" % self.source_name


class Decision(Action, metaclass=abc.ABCMeta):
    """Decision Action.

    Decision nodes correspond to Action nodes where the end result
    can be obtained using multiple methods. For instance, sources
    can be obtained either by downloading them from the store, but
    also by getting them from repositories.  Each child of Decision nodes
    represents one way to obtain the desired end result (a "choice").

    The current implementation only supports 2 choices, called
    "left" and "right".

    The DAG we first create when reading the plan includes these Decision
    nodes. Then, as part of scheduling the plan, we create a new DAG where
    all Decision nodes are removed, and only one child of each Decision
    is chosen by the scheduler, then taking the place of its Decision
    node in the DAG. Therefore, once scheduling of the DAG is complete,
    there should no longer be any Decision node left.

    It important to know that this class does not actually decide
    which choice to make. It just records some information about
    the plan, and the associated specs. It then uses that information
    to determine if a choice has actually been made, be it implicitly
    or explicitly.

    :ivar initiator: The UID of the parent node of the decision.
    :vartype initiator: str
    :ivar left_action: The first possible choice for our decision.
    :vartype left_action: Action.
    :ivar right_action: The second possible choice for our decision.
    :vartype right_action: Action.
    :ivar choice: If not None, the choice made by the plan itself.
        It can be Decision.LEFT, meaning that the plan contains
        an entry to perform the self.left_action; If set to
        Decision.RIGHT, it means the plan contains an entry to
        perform the self.right_action; and if set to BOTH, it means
        the plan contains entries to perform both self.left_action
        and self.right_action (which may or may not be an issue).
    :vartype choice: Decision.LEFT | Decision.RIGHT | Decision.BOTH
    :ivar expected_choice: If not None, the choice implicitly made
        by the way the specs involved in the decision are written.
        In practice, this attribute records the constraints we have
        in terms of the choices in the plan which are valid.
    :ivar triggers: A list of actions that, if present in a DAG
        being created while scheduling the plan (from which this
        Decision originates), causes a specific choice to be expected
        for this decision.  Each element of this list consists of a tuple
        with the uid of the Action triggering the choice, the expected
        choice, and a plan_line (if the action comes from the plan,
        otherwise None).
    :vartype triggers: list[(str, LEFT | RIGHT, str | None)]
    :ivar decision_maker: If not None, the plan_line where an entry
        in the plan is performing an action corresponding to one of
        our choices.
    :ivar: str
    """

    LEFT: Final = 0
    RIGHT: Final = 1
    BOTH: Final = 2

    def __init__(
        self, root: Action, left: Action, right: Action, choice: Optional[Choice] = None
    ):
        """Initialize a Decision instance.

        :param root: parent node
        :param left: Same as the left_action attribute.
        :param right: Same as the right_action attribute.
        :param choice: Same as the attribute.
        """
        super().__init__(uid=root.uid + ".decision", data=None)
        self.initiator = root.uid
        self.choice = choice
        self.expected_choice: Optional[Choice] = None
        self.left_action = left
        self.right_action = right
        self.triggers: List[Tuple[str, Choice, str]] = []
        self.decision_maker: Optional[str] = None

    @property
    def left(self) -> str:
        """return self.left_action.uid (this is a convenience property)."""
        return self.left_action.uid

    @property
    def right(self) -> str:
        """return self.right_action.uid (this is a convenience property)."""
        return self.right_action.uid

    def add_trigger(self, trigger: Action, decision: Choice, plan_line: str) -> None:
        """Add a trigger to self.triggers.

        See the description of the "triggers" attribute for more
        information as to what these triggers are used for.

        :param trigger: The action which, when scheduled, causes
            the given decision (which is actually more aptly
            described as a choice) to be recorded as the expected
            choice for our Decision.
        :param decision: The expected choice when the trigger Action is
            scheduled.
        :param plan_line: plan line associated with this action
        """
        self.triggers.append((trigger.uid, decision, plan_line))

    def apply_triggers(self, dag: DAG) -> None:
        """Apply triggers to the given dag.

        :param dag: a dag of scheduled actions
        """
        for uid, decision, _ in self.triggers:
            if uid in dag:
                if self.expected_choice is None:
                    self.expected_choice = decision
                elif self.expected_choice != decision:
                    self.expected_choice = Decision.BOTH

    def get_decision(self) -> Optional[str]:
        """Return uid of the choice made by the plan, or None.

        This function returns the choice made by the plan, if any.
        This means that if the plan doesn't include an action
        that implements any of the choices, we return None.
        Similarly, if the plan has actions implementing more than
        one of the choices, the plan hasn't made any decision
        because the decision is ambiguous.

        And finally, if the plan made a choice, but that choice
        contradicts this decision's expected_choice, then the decision
        is incorrect, and we also return None in that case.

        :return: None if the decision cannot be made; otherwise,
            return uid of the choice that was made.
        """
        if self.choice is None or self.choice == Decision.BOTH:
            return None
        else:
            if self.expected_choice is not None and self.expected_choice != self.choice:
                return None
            elif self.choice == Decision.LEFT:
                return self.left
            else:
                return self.right

    def get_expected_decision(self) -> Optional[str]:
        """Get expected decision.

        :return: uid of the expected action or None if no specific decision
            is expected.
        """
        if self.expected_choice == Decision.LEFT:
            return self.left
        elif self.expected_choice == Decision.RIGHT:
            return self.right
        else:
            return None

    def set_decision(self, which: Choice, decision_maker: Optional[str]) -> None:
        """Record a choice made by an action in a plan.

        This method should be caused when an entry in our plan
        performs an action corresponding to one of the choices
        for our decision. This method records which of the choices
        that action performs.

        :param which: Decision.LEFT or Decision.RIGHT
        :param decision_maker: Record who the decision maker is.
            This is typically the plan line of the action performing
            the choice being recorded.
        """
        if self.choice is None:
            self.choice = which
        elif self.choice != which:
            self.choice = Decision.BOTH
        self.decision_maker = decision_maker

    @classmethod
    @abc.abstractmethod
    def description(cls, decision: Choice) -> str:
        """Return a description of the decision (actually a choice).

        :param decision: The decision (actually a choice).
        :return: A description of the given parameter (named "decision",
            but actually a choice).
        """
        pass  # all: no cover

    def suggest_plan_fix(self, choice: Choice) -> Optional[str]:
        """Suggest a plan line that would fix the conflict.

        :param choice: Decision.LEFT or Decision.RIGHT

        :return: a line to add to the plan or None if no fix
             can be proposed
        """
        return None


class CreateSourceOrDownload(Decision):
    """Decision between creating or downloading a source package."""

    CREATE: Final = Decision.LEFT
    DOWNLOAD: Final = Decision.RIGHT

    def __init__(self, root: Action, left: CreateSource, right: DownloadSource):
        """Initialize a CreateSourceOrDownload instance.

        :param root: parent node
        :param left: first choice
        :param right: second choice
        """
        assert isinstance(left, CreateSource)
        assert isinstance(right, DownloadSource)
        super().__init__(root=root, left=left, right=right)

    @classmethod
    def description(cls, decision: Choice) -> str:
        return "CreateSource" if decision == Decision.LEFT else "DownloadSource"


class BuildOrDownload(Decision):
    """Decision between building or downloading a component."""

    BUILD: Final = Decision.LEFT
    INSTALL: Final = Decision.RIGHT

    def __init__(self, root: Install, left: Build, right: DownloadBinary):
        """Initialize a BuildOrDownload instance.

        :param root: parent node
        :param left: first choice
        :param right: second choice
        """
        assert isinstance(left, Build)
        assert isinstance(right, DownloadBinary)
        assert isinstance(root, Install)
        super().__init__(root=root, left=left, right=right)

    @classmethod
    def description(cls, decision: Choice) -> str:
        return "Build" if decision == Decision.LEFT else "DownloadBinary"

    def suggest_plan_fix(self, choice: Choice) -> str:
        action = self.left_action if choice == Decision.LEFT else self.right_action
        spec_instance = action.data

        args = [f'"{spec_instance.name}"']
        if spec_instance.qualifier:
            args.append(f'qualifier="{spec_instance.qualifier}"')
        args.append(f'build="{spec_instance.env.build.platform}"')
        if spec_instance.env.host.platform != spec_instance.env.build.platform:
            args.append(f'host="{spec_instance.env.host.platform}"')
        if spec_instance.env.target.platform != spec_instance.env.host.platform:
            args.append(f'target="{spec_instance.env.target.platform}"')

        return "anod_{}({})".format(spec_instance.kind, ", ".join(args))



from __future__ import annotations

from typing import TYPE_CHECKING
import e3.hash
import e3.log
from e3.net.http import HTTPSession
from e3.store.backends.base import ResourceInfo, Store, StoreError

if TYPE_CHECKING:
    from typing import Dict, Optional

logger = e3.log.getLogger("store.httpsimplestore")


class HTTPSimpleStoreResourceInfo(ResourceInfo):
    def __init__(self, url: str, sha: str):
        self.url = url
        self.sha = sha

    def verify(self, resource_path: str) -> bool:
        resource_sha = e3.hash.sha1(resource_path)
        if resource_sha != self.sha:
            logger.critical(
                "wrong sha for resource %s " "expecting %s got %s",
                resource_path,
                self.sha,
                resource_sha,
            )
            return False
        else:
            return True

    @property
    def uid(self) -> str:
        return self.sha


class HTTPSimpleStore(Store):
    def get_resource_metadata(
        self, query: Dict[str, str]
    ) -> HTTPSimpleStoreResourceInfo:
        """Return resource metadata directly computed from the query.

        There is no remote server involved here.
        :param query: a dict containing two keys 'sha' and 'url'. sha is the
            sha1sum of the resource and url is the remote url
        """
        if "sha" not in query or "url" not in query:
            raise StoreError('missing either "sha" or "url" in query')
        return HTTPSimpleStoreResourceInfo(query["url"], query["sha"])

    def download_resource_content(
        self, metadata: ResourceInfo, dest: str
    ) -> Optional[str]:
        """Download a resource.

        :param metadata: metadata associated with the resource to download
        :param dest: where to download the resource
        :return: the path to the downloaded resource
        """
        if TYPE_CHECKING:
            assert isinstance(metadata, HTTPSimpleStoreResourceInfo)
        with HTTPSession() as http:
            return http.download_file(metadata.url, dest, validate=metadata.verify)



"""Suprocesses management.

This module provides some functions and classes to ease spawn of
processes in blocking or non blocking mode, redirection of its stdout,
stderr and stdin. It also provides some helpers to check the process
status
"""


from __future__ import annotations

import errno
import logging
import os
import signal
import subprocess
import sys
import time
from typing import TYPE_CHECKING

import e3.env
import e3.log
from e3.os.fs import which
from e3.text import bytes_as_str


if TYPE_CHECKING:
    from typing import cast, Any, IO, List, Literal, NoReturn, Optional, Union

    CmdLine = List[str]
    AnyCmdLine = Union[List[CmdLine], CmdLine]
    STDOUT_VALUE = Literal[-1]
    PIPE_VALUE = Literal[-2]
    DEVNULL_VALUE = Literal[-3]

    # Make STDOUT subprocess constant visible in e3.os.process
    STDOUT = cast(STDOUT_VALUE, subprocess.STDOUT)
    PIPE = cast(PIPE_VALUE, subprocess.PIPE)
    DEVNULL = cast(DEVNULL_VALUE, subprocess.DEVNULL)
else:
    STDOUT = subprocess.STDOUT
    PIPE = subprocess.PIPE
    DEVNULL = subprocess.DEVNULL

logger = e3.log.getLogger("os.process")

# Special logger used for command line logging.
# This allow user to filter easily the command lines log from the rest
CMD_LOGGER_NAME = "os.process.cmdline"

cmdlogger = e3.log.getLogger(CMD_LOGGER_NAME)


# Use psutil.Popen when available to get psutil.Process properties and
# methods available in Run.internal
try:
    import psutil
    from psutil import Popen
except ImportError:  # defensive code
    from subprocess import Popen

    psutil = None


def subprocess_setup() -> None:
    """Reset SIGPIPE handler.

    Python installs a SIGPIPE handler by default. This is usually not
    what non-Python subprocesses expect.
    """
    # Set sigpipe only when set_sigpipe is True
    # This should fix HC16-020 and could be activated by default
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)  # all: no cover


def get_rlimit(platform: Optional[str] = None) -> str:
    if platform is None:
        platform = e3.env.Env().build.platform
    if platform == "x86_64-windows64":
        platform = "x86_64-windows"

    from pkg_resources import resource_filename

    return resource_filename(__name__, os.path.join("data", "rlimit-%s" % platform))


def quote_arg(arg: str) -> str:
    """Return the quoted version of the given argument.

    Returns a human-friendly representation of the given argument, but with all
    extra quoting done if necessary.  The intent is to produce an argument
    image that can be copy/pasted on a POSIX shell command (at a shell prompt).
    :param arg: argument to quote
    """
    # The empty argument is a bit of a special case, as it does not
    # contain any character that might need quoting, and yet still
    # needs to be quoted.
    if arg == "":
        return "''"

    need_quoting = (
        "|",
        "&",
        ";",
        "<",
        ">",
        "(",
        ")",
        "$",
        "`",
        "\\",
        '"',
        "'",
        " ",
        "\t",
        "\n",
        # The POSIX spec says that the following
        # characters might need some extra quoting
        # depending on the circumstances.  We just
        # always quote them, to be safe (and to avoid
        # things like file globbing which are sometimes
        # performed by the shell). We do leave '%' and
        # '=' alone, as I don't see how they could
        # cause problems.
        "*",
        "?",
        "[",
        "#",
        "~",
    )
    for char in need_quoting:
        if char in arg:
            # The way we do this is by simply enclosing the argument
            # inside single quotes.  However, we have to be careful
            # of single-quotes inside the argument, as they need
            # to be escaped (which we cannot do while still inside.
            # a single-quote string).
            arg = arg.replace("'", r"'\''")
            # Also, it seems to be nicer to print new-line characters
            # as '\n' rather than as a new-line...
            arg = arg.replace("\n", r"'\n'")
            return "'%s'" % arg
    # No quoting needed.  Return the argument as is.
    return arg


def to_cmd_lines(cmds: AnyCmdLine) -> List[CmdLine]:
    if isinstance(cmds[0], str):
        # Turn the simple command into a special case of
        # the multiple-commands case.  This will allow us
        # to treat both cases the same way.
        return [cmds]  # type: ignore
    else:
        return cmds  # type: ignore


def command_line_image(cmds: AnyCmdLine) -> str:
    """Return a string image of the given command(s).

    :param cmds: Same as the cmds parameter in the Run.__init__ method.

    This method also handles quoting as defined for POSIX shells.
    This means that arguments containing special characters
    (such as a simple space, or a backslash, for instance),
    are properly quoted.  This makes it possible to execute
    the same command by copy/pasting the image in a shell
    prompt.

    The result is expected to be a string that can be sent verbatim
    to a shell for execution.
    """
    return " | ".join(
        " ".join(quote_arg(arg) for arg in cmd) for cmd in to_cmd_lines(cmds)
    )


def enable_commands_handler(filename: str, mode: str = "a") -> logging.Handler:
    """Add a handler that log all commands launched with Run in a file.

    :param filename: path to log the commands
    :param mode: mode used to open the file (default is 'a')
    :return: the added handler
    """

    class CmdFilter(logging.Filter):
        """Keep only e3.os.process.cmdline records."""

        def filter(self, record: logging.LogRecord) -> int:
            return 1 if record.name == "e3." + CMD_LOGGER_NAME else 0

    # Here we don't attach the handler directly to the cmdline logger. Indeed
    # in class like e3.Main we do attach handlers to the root logger. In that
    # case only the handlers attached to root logger are called.
    rootlog = logging.getLogger()
    fh = logging.FileHandler(filename, mode=mode)
    fh.addFilter(CmdFilter())
    fh.setLevel(logging.DEBUG)
    rootlog.addHandler(fh)
    return fh


def disable_commands_handler(handler: logging.Handler) -> None:
    """Disable special handler for commands.

    :param handler: Handler returned by enable_commands_handler
    """
    logging.getLogger().removeHandler(handler)
    handler.flush()
    handler.close()


class Run:
    """Class to handle processes.

    :ivar cmds: The ``cmds`` argument passed to the __init__ method
        (a command line passed in a list, or a list of command lines passed as
        a list of list).
    :ivar status: The exit status. As the exit status is only meaningful after
        the process has exited, its initial value is None.  When a problem
        running the command is detected and a process does not get
        created, its value gets set to the special value 127.
    :ivar raw_out: process standard output as bytes (if instanciated with
        output = PIPE). Use self.out to get a decoded string.
    :ivar raw_err: same as raw_out but for standard error.
    :ivar pid: PID. Set to -1 if the command failed to run.
    """

    def __init__(
        self,
        cmds: AnyCmdLine,
        cwd: Optional[str] = None,
        output: Union[STDOUT_VALUE, DEVNULL_VALUE, PIPE_VALUE, str, IO, None] = PIPE,
        error: Union[STDOUT_VALUE, DEVNULL_VALUE, PIPE_VALUE, str, IO, None] = STDOUT,
        input: Union[DEVNULL_VALUE, PIPE_VALUE, str, IO, None] = None,
        bg: bool = False,
        timeout: Optional[int] = None,
        env: Optional[dict] = None,
        set_sigpipe: bool = True,
        parse_shebang: bool = False,
        ignore_environ: bool = True,
    ) -> None:
        """Spawn a process.

        :param cmds: two possibilities:
            1) a command line: a tool name and its arguments, passed
            in a list. e.g. ['ls', '-a', '.']
            2) a list of command lines (as defined in (1)): the
            different commands will be piped. This means that
            [['ps', '-a'], ['grep', 'vxsim']] will be equivalent to
            the system command line 'ps -a | grep vxsim'.
        :param cwd: directory in which the process should be executed (string
            or None). If None then current directory is used
        :param output: can be PIPE (default), a filename string, a fd on an
            already opened file, a python file object or None (for stdout).
        :param error: same as output or STDOUT, which indicates that the
            stderr data from the applications should be captured into the same
            file handle as for stdout.
        :param input: same as output
        :param bg: if True then run in background
        :param timeout: limit execution time (in seconds), None means
            unlimited
        :param env: dictionary for environment variables (e.g. os.environ)
        :param set_sigpipe: reset SIGPIPE handler to default value
        :param parse_shebang: take the #! interpreter line into account
        :param ignore_environ: Applies only when env parameter is not None.
            When set to True (the default), the only environment variables
            passed to the program are the ones provided by the env parameter.
            Otherwise, the environment passed to the program consists of the
            environment variables currently defined (os.environ) augmented by
            the ones provided in env.

        :raise OSError: when trying to execute a non-existent file.

        If you specify a filename for output or stderr then file content is
        reseted (equiv. to > in shell). If you prepend the filename with '+'
        then the file will be opened in append mode (equiv. to >> in shell)
        If you prepend the input with '|', then the content of input string
        will be used for process stdin.
        """

        def add_interpreter_command(cmd_line: CmdLine) -> CmdLine:
            """Add the interpreter defined in the #! line to cmd_line.

            If the #! line cannot be parsed, just return the cmd_line
            unchanged

            On windows, /usr/bin/env will be ignored to avoid a dependency on
            cygwin and /bin/bash & /bin/sh are replaced by $SHELL if defined.
            :param cmd_line: command line
            """
            if not parse_shebang:
                # nothing to do
                return cmd_line
            prog = which(cmd_line[0], default=None)
            if prog is None:
                # Not found. Do not modify the command line
                return cmd_line

            with open(prog) as f:
                try:
                    header = f.read()[0:2]
                except UnicodeDecodeError:
                    # unknown header - cannot decode the first two bytes
                    return cmd_line
                if header != "#!":
                    # Unknown header
                    return cmd_line
                # Header found, get the interpreter command in the first line
                f.seek(0)
                line = f.readline()
                interpreter_cmds = [
                    word.strip() for word in line[line.find("!") + 1 :].split()
                ]
                # Pass the program path to the interpreter
                if len(cmd_line) > 1:
                    cmd_line = [prog] + list(cmd_line[1:])
                else:
                    cmd_line = [prog]

                if sys.platform == "win32":  # unix: no cover
                    if interpreter_cmds[0] == "/usr/bin/env":
                        # On windows be sure that PATH is taken into account by
                        # using which. In some cases involving python
                        # interpreter, the python interpreter used to run this
                        # module has been used rather than the first one on the
                        # path.
                        interpreter_cmds[1] = which(
                            interpreter_cmds[1], default=interpreter_cmds[1]
                        )
                        return interpreter_cmds[1:] + cmd_line
                    elif (
                        interpreter_cmds[0] in ("/bin/bash", "/bin/sh")
                        and "SHELL" in os.environ
                    ):
                        return [os.environ["SHELL"]] + cmd_line
                return interpreter_cmds + cmd_line

        # First resolve output, error and input
        self.input_file = File(input, "r")
        self.output_file = File(output, "w")
        self.error_file = File(error, "w")

        self.status: Optional[int] = None
        self.raw_out = b""
        self.raw_err = b""
        self.cmds = []

        if env is not None:
            if ignore_environ:
                if sys.platform == "win32":
                    # On Windows not all environment variables can be
                    # discarded. At least SYSTEMDRIVE, SYSTEMROOT should be
                    # set. In order to be portable propagate their value in
                    # case the user does not pass them in env when
                    # ignore_environ is set to True.
                    tmp = {}
                    for var in ("SYSTEMDRIVE", "SYSTEMROOT"):
                        if var not in env and var in os.environ:
                            tmp[var] = os.environ[var]
                    tmp.update(env)
                    env = tmp
            else:
                # ignore_environ is False, so get a copy of the current
                # environment and update it with the env dictionary.
                tmp = os.environ.copy()
                tmp.update(env)
                env = tmp

        rlimit_args = []
        if timeout is not None:
            rlimit = get_rlimit()
            if os.path.exists(rlimit):
                rlimit_args = [rlimit, "%d" % timeout]
            else:
                logger.warning("cannot find rlimit at %s", rlimit)
                rlimit_args = []

        try:
            self.cmds = [add_interpreter_command(c) for c in to_cmd_lines(cmds)]
            self.cmds[0] = rlimit_args + list(self.cmds[0])

            cmdlogger.debug(
                "Run: cd %s; %s",
                cwd if cwd is not None else os.getcwd(),
                self.command_line_image(),
            )

            if len(self.cmds) == 1:
                popen_args = {
                    "stdin": self.input_file.fd,
                    "stdout": self.output_file.fd,
                    "stderr": self.error_file.fd,
                    "cwd": cwd,
                    "env": env,
                    "universal_newlines": False,
                }

                if sys.platform != "win32" and set_sigpipe:  # windows: no cover
                    # preexec_fn is no supported on windows
                    popen_args["preexec_fn"] = subprocess_setup  # type: ignore

                if sys.platform == "win32":
                    popen_args["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

                self.internal = Popen(self.cmds[0], **popen_args)

            else:
                runs: List[Popen] = []
                for index, cmd in enumerate(self.cmds):
                    if index == 0:
                        stdin = self.input_file.fd
                    else:
                        stdin = runs[index - 1].stdout

                    # When connecting two processes using a Pipe don't use
                    # universal_newlines mode. Indeed commands transmitting
                    # binary data between them will crash
                    # (e.g. gzip -dc foo.txt | tar -xf -)
                    if index == len(self.cmds) - 1:
                        stdout = self.output_file.fd
                    else:
                        stdout = subprocess.PIPE

                    popen_args = {
                        "stdin": stdin,
                        "stdout": stdout,
                        "stderr": self.error_file.fd,
                        "cwd": cwd,
                        "env": env,
                        "universal_newlines": False,
                    }

                    if sys.platform != "win32" and set_sigpipe:  # windows: no cover
                        # preexec_fn is no supported on windows
                        popen_args["preexec_fn"] = subprocess_setup  # type: ignore

                    if sys.platform == "win32":
                        popen_args[
                            "creationflags"
                        ] = subprocess.CREATE_NEW_PROCESS_GROUP

                    try:
                        runs.append(Popen(cmd, **popen_args))
                    except OSError:
                        logger.error("error when spawning %s", cmd)
                        # We have an error (e.g. file not found), try to kill
                        # all processes already started.
                        for p in runs:
                            p.terminate()
                        raise

                    self.internal = runs[-1]

        except Exception as e:  # defensive code
            self.__error(e, self.cmds)
            raise

        self.pid = self.internal.pid

        if not bg:
            self.wait()

    @property
    def out(self) -> str:
        """Process output as string.

        Attempt is done to decode as utf-8 the output. If the output is not in
        utf-8 a string representation will be returned
        (see e3.text.bytes_as_str).
        """
        return bytes_as_str(self.raw_out)

    @property
    def err(self) -> str:
        """Process error as string.

        Attempt is done to decode as utf-8 the output. If the output is not in
        utf-8 a string representation will be returned
        (see e3.text.bytes_as_str).
        """
        return bytes_as_str(self.raw_err)

    def command_line_image(self) -> str:
        """Get shell command line image of the spawned command(s).

        This just a convenient wrapper around the function of the same
        name.
        """
        return command_line_image(self.cmds)

    def close_files(self) -> None:
        """Close all file descriptors."""
        self.output_file.close()
        self.error_file.close()
        self.input_file.close()

    def __error(self, error: Exception, cmds: List[CmdLine]) -> None:
        """Set pid to -1 and status to 127 before closing files."""
        self.close_files()
        logger.error(error)

        def not_found(path: str) -> NoReturn:
            """Raise OSError.

            :param path: path of the executable
            """
            logger.error("%s not found", path)
            e3.log.debug("PATH=%s", os.environ["PATH"])
            raise OSError(
                errno.ENOENT, "No such file or directory, %s not found" % path
            )

        # Try to send an helpful message if one of the executable has not
        # been found.
        for cmd in cmds:
            if which(cmd[0], default=None) is None:
                not_found(cmd[0])

    def wait(self) -> int:
        """Wait until process ends and return its status.

        :return: exit code of the process
        """
        if self.status is not None:
            # Wait has already been called
            return self.status

        # If there is no pipe in the loop then just do a wait. Otherwise
        # in order to avoid blocked processes due to full pipes, use
        # communicate.
        if (
            self.output_file.fd != subprocess.PIPE
            and self.error_file.fd != subprocess.PIPE
            and self.input_file.fd != subprocess.PIPE
        ):
            self.status = self.internal.wait()
        else:
            tmp_input: Optional[Union[str, bytes]] = None
            if self.input_file.fd == subprocess.PIPE:
                tmp_input = self.input_file.get_command()

            if isinstance(tmp_input, str):
                tmp_input = tmp_input.encode("utf-8")

            (self.raw_out, self.raw_err) = self.internal.communicate(tmp_input)
            self.status = self.internal.returncode

        self.close_files()
        return self.status

    def poll(self) -> Optional[int]:
        """Check the process status and set self.status if available.

        This method checks whether the underlying process has exited
        or not. If it hasn't, then it just returns None immediately.
        Otherwise, it stores the process' exit code in self.status
        and then returns it.

        :return: None if the process is still alive; otherwise, returns
          the process exit status.
        """
        if self.status is not None:
            # Process is already terminated and wait been called
            return self.status

        result = self.internal.poll()

        if result is not None:
            # Process is finished, call wait to finalize it (closing handles,
            # ...)
            return self.wait()
        else:
            return None

    def kill(self, recursive: bool = True, timeout: int = 3) -> None:
        """Kill the process.

        :param recursive: if True, try to kill the complete process tree
        :param timeout: wait timeout (in seconds) after sending the kill
            signal (when recursive=True)
        """
        if recursive:
            kill_process_tree(self.internal, timeout=timeout)
        else:
            self.internal.kill()

    def interrupt(self) -> None:
        """Send SIGINT to the process, kill on Windows."""
        if sys.platform == "win32":
            self.kill()  # Ctrl-C event is unreliable on Windows
        else:
            self.internal.send_signal(signal.SIGINT)

    def is_running(self) -> bool:
        """Check whether the process is running."""
        if psutil is None:  # defensive code
            # psutil not imported, use our is_running function
            return is_running(self.pid)
        else:
            return self.internal.is_running()

    def children(self) -> List[psutil.Process]:
        """Return list of child processes (using psutil)."""
        if psutil is None:  # defensive code
            raise NotImplementedError("Run.children() require psutil")
        return self.internal.children()


class File:
    """Can be a PIPE, a file object."""

    def __init__(self, name: Any, mode: str = "r"):
        """Create a new File.

        :param name: can be PIPE, STDOUT, a filename string, an opened fd, a
            python file object, or a command to pipe (if starts with |)
        :param mode: can be 'r' or 'w' if name starts with + the mode will be
            a+
        """
        assert mode in "rw", "Mode should be r or w"
        self.fd: Union[int, IO[str]]

        self.name = name
        self.to_close = False
        if isinstance(name, str):
            # can be a pipe or a filename
            if mode == "r" and name.startswith("|"):
                self.fd = subprocess.PIPE
            else:
                if mode == "w":
                    if name.startswith("+"):
                        open_mode = "a+"
                        name = name[1:]
                    else:
                        open_mode = "w+"
                else:
                    open_mode = "r"

                self.fd = open(name, open_mode)
                if open_mode == "a+":
                    self.fd.seek(0, 2)
                self.to_close = True

        else:
            # this is a file descriptor
            self.fd = name

    def get_command(self) -> Optional[str]:
        """Return the command to run to create the pipe."""
        if self.fd == subprocess.PIPE:
            return self.name[1:]
        else:
            return None

    def close(self) -> None:
        """Close the file if needed."""
        if self.to_close:
            fd = self.fd
            fd.close()  # type: ignore


class WaitError(Exception):
    pass


def wait_for_processes(process_list: List[Run], timeout: float) -> Optional[int]:
    """Wait for several processes spawned with Run.

    :param process_list: a list of Run objects
    :param timeout: a timeout in seconds. If 0 block until a process ends.

    :return: None in case of timeout or the index in process Run corresponding
        to the first process that end
    """
    if len(process_list) == 0:
        return None

    start = time.time()
    remain = timeout

    if sys.platform == "win32":  # unix: no cover
        from e3.os.windows.process import process_exit_code, wait_for_objects

        handles = [int(p.internal._handle) for p in process_list]

        while True:
            try:
                idx = wait_for_objects(handles, remain, False)
                if idx is None:
                    return

                if process_exit_code(handles[idx]) is None:
                    # Process is still active so wait after updating timeout
                    remain = timeout - time.time() + start

                    if remain <= 0:
                        # No remaining time
                        return None
                else:
                    # Process is exiting so finalize it by calling wait
                    process_list[idx].wait()
                    return idx
            except OSError:
                raise WaitError

    else:  # windows: no cover
        import select

        # Each time a SIGCHLD signal is received write into pipe. Use
        # then select which support timeout arguments to wait.
        fd_r, fd_w = os.pipe()

        def handler(signum: int, frame: Any) -> None:
            del signum, frame
            os.write(fd_w, b"a")

        signal.signal(signal.SIGCHLD, handler)

        try:
            while remain >= 0.0 or timeout == 0:
                # Do a first check in case a SIGCHLD was emited before the
                # initialisation of the handler.
                for index, p in enumerate(process_list):
                    if p.poll() is not None:
                        return index

                # Wait for a sigchld signal. Note that select might
                # be interrupted by signals thus the loop
                select_args = [[fd_r], [], []]
                if timeout != 0:
                    select_args.append(remain)

                while True:
                    try:
                        l_r, _, _ = select.select(*select_args)  # type: ignore
                        if l_r:
                            os.read(fd_r, 1)
                        break
                    except OSError:
                        pass

                remain = timeout - time.time() + start

            logger.warning(
                "no process ended after %f seconds", time.time() - start
            )  # defensive code

        finally:
            # Be sure to remove signal handler and close pipe
            signal.signal(signal.SIGCHLD, 0)
            os.close(fd_r)
            os.close(fd_w)
    return None


def is_running(pid: int) -> bool:
    """Check whether a process with the given pid is running.

    :param pid: an integer (e.g the value of Run().pid)
    """
    if sys.platform == "win32":  # unix: no cover
        from e3.os.windows.native_api import Access, NT
        from e3.os.windows.process import process_exit_code

        handle = NT.OpenProcess(Access.PROCESS_QUERY_INFORMATION, False, pid)

        try:
            if not handle:
                return False
            return process_exit_code(handle) is None

        finally:
            NT.Close(handle)

    else:  # windows: no cover
        try:
            # We send a null signal to check the validity of pid
            os.kill(pid, 0)
        except OSError as e:
            # If the process is not found, errno will be set to ESRCH
            return e.errno != errno.ESRCH
        return True


def kill_process_tree(pid: Union[int, psutil.Process], timeout: int = 3) -> bool:
    """Kill a hierarchy of processes.

    :param pid: pid of the toplevel process
    :param timeout: wait timeout after sending the kill signal
    :return: True if all processes either don't exist or have been killed,
        False if there are some processes still alive.
    """
    if isinstance(pid, psutil.Process):
        parent_process = pid
    else:
        try:
            parent_process = psutil.Process(pid)
        except psutil.NoSuchProcess as err:
            e3.log.debug(err)
            return True

    logger.debug("kill_process_tree %s", parent_process)

    try:
        children = parent_process.children(recursive=True)
    except psutil.NoSuchProcess as err:
        e3.log.debug(err)
        return True

    all_processes = [parent_process] + children
    for p in all_processes:
        try:
            logger.debug("kill process %s (%s)", p, p.cmdline())
            p.kill()
        except psutil.NoSuchProcess:  # defensive code
            pass

    def on_terminate(p: str) -> None:
        """Log info when a process terminate."""
        logger.info("process %s killed", p)

    try:
        gone, alive = psutil.wait_procs(
            all_processes, timeout=timeout, callback=on_terminate
        )
        e3.log.debug("%d processes killed", len(gone))
        for p in alive:  # defensive code
            logger.warn("process %s survived kill()", p)

        return True
    except psutil.TimeoutExpired as err:  # defensive code
        e3.log.debug(err)
        return False



from e3.anod.context import AnodContext
from e3.anod.loader import AnodSpecRepository
from e3.anod.queries import SourceClosure, get_build_node
import os


class TestSourceClosure:

    spec_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "source_closure_specs")
    )

    def get_source_closure(self, name, expand_packages=True, other_builds=None):
        asr = AnodSpecRepository(self.spec_dir)
        ac = AnodContext(asr)
        anod_instance = ac.add_anod_action(
            name, env=ac.default_env, primitive="build"
        ).data
        anod_instance = get_build_node(anod_instance, ac, anod_instance)
        if other_builds is not None:
            for b in other_builds:
                ac.add_anod_action(b, env=ac.default_env, primitive="build")
        return SourceClosure(
            anod_instance=anod_instance, context=ac, expand_packages=expand_packages
        )

    def test_empty_source_closure(self):
        """Simple test with null source closure."""
        sc = self.get_source_closure("spec1")
        sources = sc.get_source_list()
        assert len(sources) == 0

    def test_simple_source_closure(self):
        """Simple test with one public and one private source."""
        sc = self.get_source_closure("spec2", other_builds=["spec1"])
        for key in sc.source_list:
            sc.resolve_source(source_name=key.src_name, data=key.src_name)

        sources = sc.get_source_list()
        published_sources = [src for src, publish in sources if publish]
        unpublished_sources = [src for src, publish in sources if not publish]

        assert len(sources) == 2
        assert len(published_sources) == 1 and published_sources[0] == "spec2-src"
        assert (
            len(unpublished_sources) == 1
            and unpublished_sources[0] == "spec2-internal-src"
        )

    def test_recursive_source_closure(self):
        sc = self.get_source_closure("spec3")

        for key in sc.source_list:
            sc.resolve_source(source_name=key.src_name, data=key.src_name)

        sources = sc.get_source_list()
        published_sources = [src for src, publish in sources if publish]
        unpublished_sources = [src for src, publish in sources if not publish]

        assert len(sources) == 4
        assert len(published_sources) == 1
        assert published_sources[0] == "spec3-src"
        assert len(unpublished_sources) == 3

        sc = self.get_source_closure("spec4")

        for key in sc.source_list:
            sc.resolve_source(source_name=key.src_name, data=key.src_name)

        sources = sc.get_source_list()
        published_sources = [src for src, publish in sources if publish]
        published_sources.sort()
        unpublished_sources = [src for src, publish in sources if not publish]

        assert len(sources) == 4
        assert len(published_sources) == 2
        assert published_sources == ["spec2-src", "spec4-src"]
        assert len(unpublished_sources) == 2

    def test_recursive_source_closure_with_package(self):
        sc = self.get_source_closure("spec5", expand_packages=True)

        for key in sc.source_list:
            sc.resolve_source(source_name=key.src_name, data=key.src_name)

        sources = sc.get_source_list()
        published_sources = [src for src, publish in sources if publish]
        published_sources.sort()
        unpublished_sources = [src for src, publish in sources if not publish]

        assert len(sources) == 4
        assert len(published_sources) == 2
        assert published_sources == ["spec2-src", "spec5-src"]
        assert len(unpublished_sources) == 2

        sc = self.get_source_closure("spec5", expand_packages=False)

        for key in sc.source_list:
            sc.resolve_source(source_name=key.src_name, data=key.src_name)

        for key in sc.package_list:
            if not key.track and not key.has_closure:
                print("skip %s" % key.anod_uid)
                continue
            name = key.anod_uid.split(".")[2]
            sc.resolve_package(
                key.anod_uid,
                [("%s-package-src" % name, True), ("%s-package-int-src" % name, False)],
            )

        sources = sc.get_source_list()
        published_sources = [src for src, publish in sources if publish]
        published_sources.sort()
        unpublished_sources = [src for src, publish in sources if not publish]

        assert len(sources) == 4
        assert len(published_sources) == 2
        assert published_sources == ["spec2-package-src", "spec5-src"]
        assert len(unpublished_sources) == 2



import e3.hash

import pytest


def test_hash():
    with open("to-hash.txt", "wb") as f:
        f.write(b"content\n")
    assert e3.hash.md5(f.name) == "f75b8179e4bbe7e2b4a074dcef62de95"
    assert e3.hash.sha1(f.name) == "7fe70820e08a1aac0ef224d9c66ab66831cc4ab1"
    assert (
        e3.hash.sha256(f.name)
        == "434728a410a78f56fc1b5899c3593436e61ab0c731e9072d95e96db290205e53"
    )
    with pytest.raises(e3.hash.HashError):
        e3.hash.md5("doesnotexist")



import os
import sys

from e3.fs import echo_to_file, mkdir, rm
from e3.os.fs import touch, unixpath
from e3.os.process import Run
from e3.vcs.svn import SVNError, SVNRepository

import pytest


def file_url(path, unix=False):
    if sys.platform == "win32":
        if len(path) > 1 and path[1] == ":":
            # svn info returns the URL with an uppercase letter drive
            path = path[0].upper() + path[1:]
        if unix:
            return "file://" + unixpath(path)
        else:
            return "file:///" + path.replace("\\", "/")
    else:
        return "file://" + path


@pytest.mark.svn
def test_svn_repo():
    cwd = os.getcwd()

    # --- create local project
    project_path = os.path.join(cwd, "test_project")
    mkdir(project_path)
    mkdir(os.path.join(project_path, "trunk"))
    hello_relative_path = os.path.join("trunk", "hello.txt")
    hello_path = os.path.join(project_path, hello_relative_path)
    echo_to_file(hello_path, "hello")

    # --- create a SVN repository from that project
    repos_path = os.path.join(cwd, "repos")
    project_url = SVNRepository.create(repo_path=repos_path)
    project_url = project_url + "/Test_Project"
    p = Run(
        ["svn", "import", unixpath(project_path), project_url, "-m", "initial import"]
    )
    assert p.status == 0, p.out

    # --- checkout project into working dir A
    working_copy_a_path = os.path.join(cwd, "working_copy_a")
    svn_a = SVNRepository(working_copy_a_path)
    with pytest.raises(SVNError):
        svn_a.update()
    with pytest.raises(SVNError):
        svn_a.update(url=file_url("bad_url"))
    local_change = svn_a.update(project_url)
    assert local_change
    local_change = svn_a.update()
    assert not local_change
    # verify the content of the working dir A and its revision
    assert svn_a.url == project_url
    assert os.path.exists(
        os.path.join(working_copy_a_path, hello_relative_path)
    ), "checkout failed"
    assert svn_a.current_revision == "1"
    # modify the working dir, commit the change,
    # update the working dir and verify the new current revision
    echo_to_file(os.path.join(working_copy_a_path, hello_relative_path), "bye")
    svn_a.svn_cmd(["commit", "-m", "modify hello"])
    svn_a.update()
    assert svn_a.current_revision == "2"
    svn_a.update(revision="1")
    assert svn_a.current_revision == "1"
    with pytest.raises(SVNError):
        svn_a.update(revision="404")

    # make local changes in the working dir B before updating it
    working_copy_b_path = os.path.join(cwd, "working_copy_b")
    svn_b = SVNRepository(working_copy_b_path)
    svn_b.update(project_url)
    foo_path = os.path.join(working_copy_b_path, "trunk", "foo")
    touch(foo_path)
    hello_b_path = os.path.join(working_copy_b_path, hello_relative_path)
    echo_to_file(hello_b_path, "kitty")
    local_change = svn_b.update()
    assert local_change
    assert os.path.exists(foo_path)
    with open(hello_b_path) as f:
        assert "kitty" in f.read()
    # update and cancel all changes
    svn_b.update(force_and_clean=True)
    assert not os.path.exists(foo_path)
    with open(hello_b_path) as f:
        assert "bye" in f.read()

    # checkout into an existing path (not versioned)
    working_copy_c_path = os.path.join(cwd, "working_copy_c")
    svn_c = SVNRepository(working_copy_c_path)
    touch(working_copy_c_path)
    with pytest.raises(SVNError) as err:
        svn_c.update(url=project_url)
    assert "not empty" in str(err)
    rm(working_copy_c_path)
    mkdir(working_copy_c_path)

    bar_path = os.path.join(working_copy_c_path, "bar")
    touch(bar_path)
    # verify failures without the force option
    with pytest.raises(SVNError) as err:
        svn_c.update(url=project_url)
    assert "not empty" in str(err)
    touch(os.path.join(working_copy_c_path, ".svn"))
    with pytest.raises(SVNError):
        svn_c.update(url=project_url)
    svn_c.update(url=project_url, force_and_clean=True)
    # verify that the working dir C is clean
    assert not os.path.exists(bar_path)

    # modify a working dir and update it with a new project
    svn_d = SVNRepository(working_copy_c_path)
    touch(bar_path)
    svn_d.update()  # update with the last URL used for this dir
    svn_d.update(url=project_url)  # update with the same URL
    assert os.path.exists(bar_path)
    project2_url = project_url + "2"
    p = Run(
        ["svn", "import", unixpath(project_path), project2_url, "-m", "initial import"]
    )
    assert p.status == 0, p.out
    with pytest.raises(SVNError) as err:
        svn_d.update(url=project2_url)  # update with new URL
    assert "not empty" in str(err)
    svn_d.update(url=project2_url, force_and_clean=True)
    assert svn_d.url == project2_url



"""Extensions to the standard Python logging system."""

from __future__ import annotations
from dataclasses import dataclass

import logging
import re
import sys
import time
import json
from typing import TYPE_CHECKING, ClassVar

from colorama import Fore, Style

from tqdm import tqdm

from e3.config import ConfigSection

if TYPE_CHECKING:
    from typing import (
        Any,
        IO,
        Optional,
        Iterator,
        Sequence,
        TextIO,
        Union,
        List,
        Tuple,
        Mapping,
    )
    from logging import _ExcInfoType


@dataclass
class LogConfig(ConfigSection):
    title: ClassVar[str] = "log"

    pretty: bool = True
    stream_fmt: str = "%(levelname)-8s %(message)s"
    file_fmt: str = "%(asctime)s: %(name)-24s: %(levelname)-8s %(message)s"


log_config = LogConfig.load()


# Default output stream (sys.stdout by default, or a file descriptor if
# activate() is called with a filename.
default_output_stream: Union[TextIO, IO[str]] = sys.stdout

# If sys.stdout is a terminal then enable "pretty" output for user
# This includes progress bars and colors
if sys.stdout.isatty():  # all: no cover (not used in production!)
    pretty_cli = log_config.pretty
else:
    pretty_cli = False

console_logs: Optional[str] = None


class JSONFormatter(logging.Formatter):
    """Logging formatter for creating JSON logs.

    It will print some standard attributes defined in STD_ATTR
    plus application extra attributes defined in _extra_attr
    """

    # standard attributes that will always be printed
    STD_ATTR = ["asctime", "levelname", "name", "message", "module", "exc_text"]
    # custom attributes
    _extra_attr: List[str] = ["anod_uui"]

    def __init__(
        self,
        date_fmt: Optional[str] = None,
        context: Optional[Mapping[str, Any]] = None,
    ):
        """Initialize formatter with context.

        :param date_fmt: see logging module
        :param context: dict to add context information to log records
        """
        # We need to pass fmt and datefmt parameters for
        # asctime atribute to be created
        super(JSONFormatter, self).__init__(fmt="%(asctime)s", datefmt=date_fmt)

        if context is None:
            context = {}
        self.context = context

    def format(self, record: logging.LogRecord) -> str:
        """convert record into JSON."""
        # Parent's format is called in order to setup additional attributes
        super(JSONFormatter, self).format(record)

        json_record = {
            attr: getattr(record, attr, None)
            for attr in self.STD_ATTR + list(self._extra_attr)
        }
        # we add context information
        json_record.update(self.context)
        # we delete empty values
        json_record = {attr: val for attr, val in json_record.items() if val}

        return json.dumps(json_record)


class E3LoggerAdapter(logging.LoggerAdapter):
    """LoggerAdapter to add custom keywords."""

    def process(self, msg: Any, kwargs: Any) -> Tuple[Any, Any]:
        """Allow to handle extra parameter.

        It is called by super method log. It is overwritten here because
        the standard process method will get rid of extra attribute
        """
        return msg, kwargs

    def log(
        self, level: int, msg: Any, *args: Any, anod_uui: int = 0, **kwargs: Any
    ) -> None:
        """Integrate additional keywords using standard interface.

        :param level: see logging module
        :param args: see logging module
        :param anod_uui: Anod UUI
        :param kwargs: other parameter supported by std logger._log method
        """
        extra_attrs = {"anod_uui": anod_uui}
        extra = kwargs.setdefault("extra", {})
        # we use the standard 'extra' parameter to pass additional keywords
        extra.update(extra_attrs)
        super(E3LoggerAdapter, self).log(level, msg, *args, **kwargs)

    def info(self, msg: Any, *args: Any, anod_uui: int = 0, **kwargs: Any) -> None:
        """Wrap standard logger.info method.

        It allows adding extra keyword parameters
        """
        self.log(logging.INFO, msg, *args, anod_uui=anod_uui, **kwargs)

    def debug(self, msg: Any, *args: Any, anod_uui: int = 0, **kwargs: Any) -> None:
        """Wrap standard logger.debug method.

        It allows adding extra keyword parameters
        """
        self.log(logging.DEBUG, msg, *args, anod_uui=anod_uui, **kwargs)

    def warning(self, msg: Any, *args: Any, anod_uui: int = 0, **kwargs: Any) -> None:
        """Wrap standard logger.warning method.

        It allows adding extra keyword parameters
        """
        self.log(logging.WARNING, msg, *args, anod_uui=anod_uui, **kwargs)

    def error(self, msg: Any, *args: Any, anod_uui: int = 0, **kwargs: Any) -> None:
        """Wrap standard logger.error method.

        It allows adding extra keyword parameters
        """
        self.log(logging.ERROR, msg, *args, anod_uui=anod_uui, **kwargs)

    def critical(self, msg: Any, *args: Any, anod_uui: int = 0, **kwargs: Any) -> None:
        """Wrap of standard logger.critical method.

        It allows adding extra keyword parameters
        """
        self.log(logging.CRITICAL, msg, *args, anod_uui=anod_uui, **kwargs)

    def exception(
        self,
        msg: Any,
        *args: Any,
        exc_info: _ExcInfoType = True,
        anod_uui: int = 0,
        **kwargs: Any,
    ) -> None:
        """Wrap standard logger.exception method.

        It allows adding extra keyword parameters
        """
        self.log(
            logging.ERROR, msg, *args, exc_info=exc_info, anod_uui=anod_uui, **kwargs
        )


def progress_bar(it: Union[Iterator, Sequence], **kwargs: Any) -> tqdm:
    """Create a tqdm progress bar.

    :param it: an interator
    :param kwargs: see tqdm documentation
    :return: a tqdm progress bar iterator
    """
    if pretty_cli:  # all: no cover
        return tqdm(it, file=sys.stderr, **kwargs)
    else:
        # When pretty cli is disabled return a progress bar that do nothing.
        # returning just the iterator will break calls to tqdm method
        # otherwise.
        return tqdm(it, disable=True, file=sys.stderr, **kwargs)


__null_handler_set = set()


class TqdmHandler(logging.StreamHandler):  # all: no cover
    """Logging handler when used when progress bars are enabled."""

    # Color the log status at the beginning of most log lines
    color_subst = (
        (re.compile(r"^(DEBUG)"), Fore.CYAN),
        (re.compile(r"^(INFO)"), Style.DIM),
        (re.compile(r"^(WARNING)"), Fore.YELLOW),
        (re.compile(r"^(ERROR)"), Fore.RED),
        (re.compile(r"^(CRITICAL)"), Fore.RED + Style.BRIGHT),
    )

    def __init__(self) -> None:
        logging.StreamHandler.__init__(self)

    def emit(self, record: logging.LogRecord) -> None:
        msg = self.format(record)

        # Handle logging on several lines: indent all lines after the first one
        # to be aligned with the first one.
        msg_first_line = msg.split("\n")[0]
        msg = msg.replace(
            "\n", "\n_" + " " * (len(msg_first_line) - len(record.message) - 1)
        )

        # Add color
        for reg, color in self.color_subst:
            msg = re.sub(reg, color + r"\1" + Fore.RESET + Style.RESET_ALL, msg)

        tqdm.write(msg, file=sys.stderr)


def getLogger(name: Optional[str] = None, prefix: str = "e3") -> E3LoggerAdapter:
    """Get a logger with a default handler doing nothing.

    Calling this function instead of logging.getLogger will avoid warnings
    such as::

        'No handler could be found for logger...'

    :param name: logger name, if not specified return the root logger
    :param prefix: application prefix, will be prepended to the name
    """

    class NullHandler(logging.Handler):
        """Handler doing nothing."""

        def emit(self, _: logging.LogRecord) -> None:
            pass

    logger = logging.getLogger(f"{prefix}.{name}")

    if prefix not in __null_handler_set:
        # Make sure that the root logger has at least an handler attached to
        # it to avoid warnings.
        logging.getLogger(prefix).addHandler(NullHandler())
        __null_handler_set.add(prefix)
    return E3LoggerAdapter(logger, {})


def add_log_handlers(
    level: int,
    log_format: str,
    datefmt: Optional[str] = None,
    filename: Optional[str] = None,
    set_default_output: bool = True,
    json_format: bool = False,
) -> None:
    """Add log handlers using GMT.

    :param level: set the root logger level to the specified level
    :param log_format: format stream for the log handler
    :param datefmt: date/time format for the log handler
    :param filename: use of a FileHandler, using the specified filename,
        instead of a StreamHandler. Set default_output_stream to write in this
        file.
    """
    global default_output_stream
    handler: Union[TqdmHandler, logging.StreamHandler, logging.FileHandler]
    fmt: Union[logging.Formatter, JSONFormatter]

    if filename is None:
        if pretty_cli:  # all: no cover
            handler = TqdmHandler()
        else:
            handler = logging.StreamHandler()
    else:
        handler = logging.FileHandler(filename)
        if set_default_output:
            default_output_stream = handler.stream

    if json_format:
        fmt = JSONFormatter(datefmt, {"context": console_logs})
    else:
        fmt = logging.Formatter(log_format, datefmt)

    fmt.converter = time.gmtime  # type: ignore
    handler.setFormatter(fmt)

    handler.setLevel(level)
    logging.getLogger("").addHandler(handler)


def activate(
    stream_format: str = log_config.stream_fmt,
    file_format: str = log_config.file_fmt,
    datefmt: Optional[str] = None,
    level: int = logging.INFO,
    filename: Optional[str] = None,
    e3_debug: bool = False,
    json_format: bool = False,
) -> None:
    """Activate default E3 logging.

    :param level: set the root logger level to the specified level
    :param datefmt: date/time format for the log handler
    :param stream_format: format string for the stream handler
    :param file_format: format string for the file handler
    :param filename: redirect logs to a file in addition to the StreamHandler
    :param e3_debug: activate full debug of the e3 library
    """
    # By default do not filter anything. What is effectively logged
    # will be defined by setting/unsetting handlers
    logging.getLogger("").setLevel(logging.DEBUG)
    if console_logs:
        stream_format = f"{console_logs}: {file_format}"

    # Set logging handlers
    add_log_handlers(
        level=level, log_format=stream_format, datefmt=datefmt, json_format=json_format
    )

    # Log to a file if necessary
    if filename is not None:
        add_log_handlers(
            level=min(level, logging.DEBUG),
            log_format=file_format,
            datefmt=datefmt,
            filename=filename,
            json_format=json_format,
        )

    if e3_debug:
        getLogger("debug").setLevel(logging.DEBUG)


# Provide a logger than will provides full debug information when a program
# using e3.main.Main is called with -v -v

e3_debug_logger = getLogger("debug")
e3_debug_logger.setLevel(logging.CRITICAL + 1)

debug = e3_debug_logger.debug



from __future__ import annotations

import abc
import threading
from collections import namedtuple
from datetime import datetime
from typing import TYPE_CHECKING

import e3.log
from e3.anod.status import ReturnValue
from e3.os.process import Run


if TYPE_CHECKING:
    from typing import Any, Callable, List, Optional
    from e3.job.scheduler import Scheduler

logger = e3.log.getLogger("job")


JobTimingInfo = namedtuple("JobTimingInfo", ["start_time", "stop_time", "duration"])


class Job(metaclass=abc.ABCMeta):
    """Handle a single Job.

    :ivar slot: number associated with the job during its execution. At a
        given time only one job have a given slot number.
    :vartype slot: int
    :ivar start_time: time at which job execution started or None if never
        started
    :vartype start_time: datetime.datetime
    :ivar stop_time: time at which job execution ended or None if either
        the job was never run or if the job is still running
    :vartype stop_time: datetime.datetime
    :ivar should_skip: indicator for the scheduler that the job should not
        be executed
    :vartype: bool
    :ivar interrupted: if True it means that job has been interrupted. Can
        be consequence of timeout or Ctrl-C pressed
    :vartype: bool
    :ivar queue_name: name of the queue in which the job has been placed
    :vartype: str
    :ivar tokens: number of tokens (i.e: resources) consumed during the job
        execution
    :vartype: int
    :ivar index: global index indicating the order in which jobs have been
        created. The index is used to implement the default ordering function
        needed to sort Jobs. In the context of ``e3.job.scheduler.Scheduler``
        this means that by default jobs that are created first will have a
        higher priority.
    :vartype: int
    """

    lock = threading.RLock()
    index_counter = 0

    def __init__(self, uid: str, data: Any, notify_end: Callable[[str], None]):
        """Initialize worker.

        :param uid: unique work identifier
        :param data: work data
        :param notify_end: function that takes the job uid as parameter. The
            function is called whenever the job finish. The function is
            provided by the scheduler.
        """
        self.uid = uid
        self.data = data
        self.notify_end = notify_end
        self.slot = 1
        self.handle: Optional[threading.Thread] = None
        self.thread = None
        self.__start_time: Optional[datetime] = None
        self.__stop_time: Optional[datetime] = None
        self.should_skip = False
        self.interrupted = False
        self.queue_name = "default"
        self.tokens = 1
        with self.lock:
            self.index = Job.index_counter
            Job.index_counter += 1

    @property
    def priority(self) -> int:
        """Return job priority.

        This is used in ``e3.job.scheduler.Scheduler``.
        """
        return 0

    def record_start_time(self) -> None:
        """Log the starting time of a job."""
        with self.lock:
            self.__start_time = datetime.now()

    def record_stop_time(self) -> None:
        """Log the stopping time of a job."""
        with self.lock:
            self.__stop_time = datetime.now()

    @property
    def timing_info(self) -> JobTimingInfo:
        """Retrieve some job's timing information.

        :return: a JobTimingInfo object
        """
        with self.lock:
            start = self.__start_time
            stop = self.__stop_time
        if start is None:
            duration = 0.0
        else:
            duration = ((stop or datetime.now()) - start).total_seconds()
        return JobTimingInfo(start, stop, duration)

    def start(self, slot: int) -> None:
        """Launch the job.

        :param slot: slot number
        """

        def task_function() -> None:
            self.record_start_time()
            try:
                with self.lock:
                    interrupted = self.interrupted
                if interrupted:  # defensive code
                    logger.debug("job %s has been cancelled", self.uid)
                else:
                    self.run()
            finally:
                self.record_stop_time()
                self.notify_end(self.uid)

        self.slot = slot
        self.handle = threading.Thread(target=task_function, name=self.uid)
        self.handle.start()

    @abc.abstractmethod
    def run(self) -> None:
        """Job activity."""
        pass  # all: no cover

    @property
    def status(self) -> ReturnValue:
        """Return he job's status.

        This is made a property because users of this class should not
        be allowed to set or change it value. The job's status is ...
        a property of the job!
        """
        return ReturnValue.success

    def interrupt(self) -> bool:
        """Interrupt current job.

        :return: True if interrupted, False if already interrupted
        """
        with self.lock:
            previous_state = self.interrupted
            self.interrupted = True
        return not previous_state

    def on_start(self, scheduler: Scheduler) -> None:
        """Call whenever a job is started.

        This allow the user to do some logging on job startup
        """
        pass

    def on_finish(self, scheduler: Scheduler) -> None:
        """Call whenever a job is finished.

        This allow the user to do some logging on job termination
        """
        pass


class EmptyJob(Job):
    """A job which does nothing."""

    def __init__(
        self,
        uid: str,
        data: Any,
        notify_end: Callable[[str], None],
        status: ReturnValue = ReturnValue.force_skip,
    ) -> None:
        """Initialize the EmptyJob.

        :param status: The job's status.
        """
        super().__init__(uid, data, notify_end)
        self.should_skip = True
        self.__status = status

    def run(self) -> None:
        pass

    @property
    def status(self) -> ReturnValue:
        """See Job.status' description."""
        return self.__status


class ProcessJob(Job, metaclass=abc.ABCMeta):
    """Specialized version of Job that spawn processes.

    :ivar proc_handle: None when an object of this class is initialized.
        An e3.os.process.Run object after the "run" method is called.
    :vartype proc_handle: e3.os.process.Run | None
    """

    def __init__(self, uid: str, data: Any, notify_end: Callable[[str], None]):
        super().__init__(uid, data, notify_end)
        self.proc_handle: Optional[Run] = None

        # Detect spawn issue to avoid returning "notready"
        # and creating a loop
        self.__spawn_error = False

    def run(self) -> None:
        """Run the job."""
        cmd_options = self.cmd_options

        # Do non blocking spawn followed by a wait in order to have
        # self.proc_handle set. This allows support for interrupt.
        cmd_options["bg"] = True
        with self.lock:
            if self.interrupted:  # defensive code
                logger.debug("job %s has been cancelled", self.uid)
                return
            try:
                cmdline = self.cmdline
                assert cmdline is not None, "cmdline cannot be None"

                proc_handle = Run(cmdline, **cmd_options)
                self.proc_handle = proc_handle
            except Exception:
                logger.exception("error when spawing job %s", self.uid)
                self.__spawn_error = True
                return
        proc_handle.wait()
        logger.debug(
            "job %s status %s (pid:%s)", self.uid, proc_handle.status, proc_handle.pid
        )

    @property
    def status(self) -> ReturnValue:
        """See Job.status' description."""
        if self.__spawn_error:
            return ReturnValue.failure
        elif self.proc_handle is None:
            return ReturnValue.notready
        else:
            try:
                return ReturnValue(self.proc_handle.status)
            except ValueError:
                logger.exception(
                    "job %s returned an unknown status %s",
                    self.uid,
                    self.proc_handle.status,
                )
                return ReturnValue.failure

    @abc.abstractproperty
    def cmdline(self) -> List[str]:
        """Return the command line of the process to be spawned.

        :return: the command line
        """
        pass  # all: no cover

    @property
    def cmd_options(self) -> dict:
        """Process options.

        Important note: don't use PIPE for output or error parameters this can
        cause locking error in case the process is interrupted. The default
        redirect output and error to the console.

        The pipe behavior can easily be emulated by writing to a file and
        modifying the run method to read the file content when the process
        finish.

        :return: options for e3.os.process.Run as a dict
        """
        return {"output": None}

    def interrupt(self) -> bool:
        """Kill running process tree."""
        if super().interrupt():
            if self.proc_handle is None:  # defensive code
                logger.debug("cancel job %s", self.uid)
            elif self.proc_handle.is_running():
                logger.debug("kill job %s", self.uid)
                self.proc_handle.kill(recursive=True)
            else:  # defensive code
                logger.debug("cannot interrupt, job %s has finished", self.uid)
        return True



"""Helpers classes and functions for ANOD."""


from __future__ import annotations

import io
import os
import re
from typing import TYPE_CHECKING

import e3.log
from e3.anod.spec import Anod, parse_command
from e3.os.fs import unixpath

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Tuple, Union
    from e3.os.process import Run

log = e3.log.getLogger("anod.helpers")


class Make:
    """Wrapper around GNU Make."""

    def __init__(
        self,
        anod_instance: Anod,
        makefile: Optional[str] = None,
        exec_dir: Optional[str] = None,
        jobs: Optional[int] = None,
    ):
        """Initialize a Make object.

        :param anod_instance: an Anod instance
        :param makefile: the Makefile to use
        :param exec_dir: path to the directory from where the make should be
            called. If None use the anod instance buildspace build dir.
        :param jobs: number of jobs to run in parallel
        """
        self.anod_instance = anod_instance
        self.exec_dir = exec_dir

        if self.exec_dir is None:
            self.exec_dir = self.anod_instance.build_space.build_dir
        self.makefile = makefile
        self.jobs = jobs
        if jobs is None:
            self.jobs = anod_instance.jobs
        self.var_list = {}  # type: Dict[str, str]
        self.default_target = None  # type: Optional[str]

    def set_var(self, name: str, value: Union[str, List[str]]) -> None:
        """Set a Make variable.

        :param name: name of the variable
        :param value: value of the variable, can be a string or a list
            if it's the list, it will be stored in a string with each
            value separated by a space character
        """
        if isinstance(value, str):
            self.var_list[name] = value
        else:
            # Assume we get a list
            self.var_list[name] = " ".join(value)

    def set_default_target(self, target: str) -> None:
        """Set default make target.

        :param target: the target name to use if __call__ is called with
            target=None
        """
        self.default_target = target

    def __call__(
        self,
        target: Optional[str] = None,
        jobs: Optional[int] = None,
        exec_dir: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> Run:
        """Call a make target.

        :param target: the target to use (use default_target if None)
        :param jobs: see __init__ documentation
        :param exec_dir: see __init__ documentation
        :param timeout: timeout to pass to ex.Run
        """
        cmdline = self.cmdline(target, jobs, exec_dir, timeout=timeout)
        cmd = cmdline["cmd"]
        options = cmdline["options"]
        return self.anod_instance.shell(*cmd, **options)  # type: ignore

    def cmdline(
        self,
        target: Optional[Union[str, List[str]]] = None,
        jobs: Optional[int] = None,
        exec_dir: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> Dict[str, Union[List[str], dict]]:
        """Return the make command line.

        :param target: optional target or list of targets to use
            (use default_target if None)
        :param jobs: see __init__ documentation
        :param exec_dir: see __init__ documentation
        :param timeout: timeout to pass to ex.Run

        :return: a dictionary with the following keys
           - cmd: containing the command line to pass to gnatpython.ex.Run
           - options: options to pass to gnatpython.ex.Run
        """
        cmd_arg_list = ["make"]

        if self.makefile is not None:
            cmd_arg_list += ["-f", unixpath(self.makefile)]

        cmd_arg_list += ["-j", "%s" % str(jobs) if jobs is not None else str(self.jobs)]

        for key in self.var_list:
            cmd_arg_list.append("{}={}".format(key, self.var_list[key]))

        if target is None:
            target = self.default_target

        if target is not None:
            if isinstance(target, list):
                cmd_arg_list += target
            else:
                cmd_arg_list.append(target)

        options = {"cwd": exec_dir or self.exec_dir, "timeout": timeout}

        return {
            "cmd": parse_command(
                command=cmd_arg_list, build_space=self.anod_instance.build_space
            ),
            "options": options,
        }


class Configure:
    """Wrapper around ./configure."""

    def __init__(
        self,
        anod_instance: Anod,
        src_dir: Optional[str] = None,
        exec_dir: Optional[str] = None,
        auto_target: bool = True,
    ):
        """Initialize a Configure object.

        :param anod_instance: an Anod instance
        :param src_dir: path to the directory containing the project sources.
            If None then use the anod_instance buildspace source dir.
        :param exec_dir: path to the directory from where the configure should
            be called. If None then use the anod_instance buildspace build
            dir.
        :param auto_target: if True, automatically pass --target, --host and
            --build
        """
        self.anod_instance = anod_instance
        if src_dir is None:
            self.src_dir = self.anod_instance.build_space.src_dir
        else:
            self.src_dir = src_dir
        self.exec_dir = exec_dir
        if self.exec_dir is None:
            self.exec_dir = self.anod_instance.build_space.build_dir
        self.args = []  # type: List[str]

        # Value of the --target, --host and --build arguments
        self.target = None
        self.host = None
        self.build = None

        if auto_target:
            e = anod_instance.env
            if e.is_canadian:
                self.target = e.target.triplet
                self.host = e.host.triplet
                self.build = e.build.triplet
            elif e.is_cross:
                self.target = e.target.triplet
                self.build = e.build.triplet
            else:
                self.build = e.target.triplet

        self.env = {}  # type: dict

    def add(self, *args: str) -> None:
        """Add configure options.

        :param args: list of options to pass when calling configure
        """
        self.args += args

    def add_env(self, key: str, value: str) -> None:
        """Set environment variable when calling configure.

        :param key: environment variable name
        :param value: environment variable value
        """
        self.env[key] = value

    def cmdline(self) -> dict:
        """Return the configure command line.

        :return: a dictionary with the following keys
           - cmd: containing the command line to pass to gnatpython.ex.Run
           - options: options to pass to gnatpython.ex.Run

        If CONFIG_SHELL environment variable is set, the configure will be
        called with this shell.
        """
        cmd = []
        if "CONFIG_SHELL" in os.environ:
            cmd.append(os.environ["CONFIG_SHELL"])

        # Compute the relative path for configure
        configure_path = unixpath(
            os.path.relpath(os.path.join(self.src_dir, "configure"), self.exec_dir)
        )

        # In case the configure is run from its location ensure to
        # add ./ as . is not necessary in PATH.
        if configure_path == "configure":
            configure_path = "./configure"
        cmd += [configure_path]
        cmd += self.args

        if self.target is not None:
            cmd.append("--target=" + self.target)

        if self.host is not None:
            cmd.append("--host=" + self.host)

        if self.build is not None:
            cmd.append("--build=" + self.build)

        cmd_options = {"cwd": self.exec_dir, "ignore_environ": False, "env": self.env}

        return {
            "cmd": parse_command(
                command=cmd, build_space=self.anod_instance.build_space
            ),
            "options": cmd_options,
        }

    def __call__(self) -> Run:
        cmdline = self.cmdline()
        cmd = cmdline["cmd"]
        options = cmdline["options"]
        return self.anod_instance.shell(*cmd, **options)


def text_replace(
    filename: str, pattern: List[Tuple[Union[bytes, str], Union[bytes, str]]]
) -> List[int]:
    """Replace patterns in a file.

    :param filename: file path
    :param pattern: list of tuple (pattern, replacement)

    Do not modify the file if no substitution is done. Note that substitutions
    are applied sequentially (order provided by the list `pattern`) and this
    is done line per line.

    :return: the number of substitution performed for each pattern
    """
    output = io.BytesIO()
    nb_substitution = [0 for _ in pattern]
    with open(filename, "rb") as f:
        for line in f:
            for pattern_index, (regexp, replacement) in enumerate(pattern):
                if isinstance(replacement, str):
                    replacement = replacement.encode("utf-8")
                if isinstance(regexp, str):
                    regexp = regexp.encode("utf-8")
                line, count = re.subn(regexp, replacement, line)
                if count:
                    nb_substitution[pattern_index] += count
            output.write(line)
    if any(nb for nb in nb_substitution):
        # file changed, update it
        with open(filename, "wb") as f:
            f.write(output.getvalue())
    output.close()
    return nb_substitution



import os

import e3.anod.error
import e3.anod.package
from e3.fs import mkdir, rm
from e3.os.fs import touch

import pytest


def test_source_builder_default_prepare_src():
    """Test SourceBuilder with the default prepare_src function."""
    sb = e3.anod.package.SourceBuilder(
        name="a-src", fullname=lambda: "a-src.tgz", checkout=["a-git"]
    )

    assert sb.fullname() == "a-src.tgz"

    current_dir = os.getcwd()
    a_wd = os.path.join(current_dir, "a_wd")

    # create a fake working dir of a-git
    mkdir(a_wd)
    touch(os.path.join(a_wd, "a_file"))

    a_dest = os.path.join(current_dir, "a_dest")
    mkdir(a_dest)

    sb.prepare_src(repos={"a-git": {"working_dir": a_wd}}, dest=a_dest)

    assert os.path.exists(os.path.join(a_dest, "a_file"))

    # Check that this is working only when we have one repo
    for checkout in ([], ["a-git", "b-git"]):
        sb_err = e3.anod.package.SourceBuilder(
            name="a-src", fullname=lambda: "a-src.tgz", checkout=checkout
        )

        with pytest.raises(e3.anod.error.SpecError):
            assert sb_err.prepare_src

    # Check ThirdParty prepare_src is None
    tp = e3.anod.package.ThirdPartySourceBuilder(name="unmanaged-src")
    assert tp.prepare_src is None


def test_source_builder_custom_prepare_src():
    """Test SourceBuilder with a custom prepare_src function."""

    def prepare_src(repos, dest):
        touch(os.path.join(dest, "my_generated_source_file"))

    sb = e3.anod.package.SourceBuilder(
        name="b-src",
        fullname=lambda: "b-src.tgz",
        checkout=["b-git"],
        prepare_src=prepare_src,
    )

    sb.prepare_src(None, os.getcwd())
    assert os.path.exists("my_generated_source_file")


def test_apply_patch():
    """Test SourceBuilder.apply_patch handling."""
    sb = e3.anod.package.SourceBuilder(
        name="d-src", fullname=lambda: "d-src.tgz", checkout=["d-git"]
    )

    with open("my_patch", "w") as f:
        f.write(
            "--- a_file\t2017-04-11 16:34:44.000000000 +0200\n"
            "+++ a_file\t2017-04-11 16:34:40.000000000 +0200\n"
            "@@ -0,0 +1 @@\n"
            "+new line\n"
        )

    current_dir = os.getcwd()
    touch("a_file")
    sb.apply_patch(None, "my_patch", current_dir)

    with open("a_file") as f:
        assert f.read().strip() == "new line"

    sb = e3.anod.package.SourceBuilder(
        name="d-src",
        fullname=lambda: "d-src.tgz",
        prepare_src=lambda x, y: None,
        checkout=["d-git", "e-git"],
    )

    with pytest.raises(e3.anod.error.AnodError) as err:
        sb.apply_patch(None, "my_patch", current_dir)
    assert "no apply_patch" in str(err)

    sb = e3.anod.package.SourceBuilder(
        name="d-src",
        fullname=lambda: "d-src.tgz",
        prepare_src=lambda x, y: None,
        checkout=["d-git", "e-git"],
        apply_patch=lambda x, y, z: 42,
    )

    assert sb.apply_patch(None, None, None) == 42

    # Thirdparty source builders set the default patch command by default

    rm("a_file")
    touch("a_file")
    tsb = e3.anod.package.ThirdPartySourceBuilder(name="third-src.tgz")
    tsb.apply_patch(None, "my_patch", current_dir)

    with open("a_file") as f:
        assert f.read().strip() == "new line"



from __future__ import annotations

import os

from typing import TYPE_CHECKING

import e3.error
import e3.log
from e3.anod.driver import AnodDriver
from e3.anod.queries import get_source_builder
from e3.anod.status import ReturnValue
from e3.fs import mkdir, sync_tree
from e3.job import Job
from e3.job.scheduler import Scheduler
from e3.vcs.git import GitRepository

if TYPE_CHECKING:
    from typing import Callable, Dict, FrozenSet, Literal
    from e3.anod.action import Action
    from e3.anod.sandbox import SandBox
    from e3.anod.loader import AnodSpecRepository
    from e3.collection.dag import DAG
    from e3.store.backends.base import Store

logger = e3.log.getLogger("electrolyt.exec")
STATUS = ReturnValue


class ElectrolytJob(Job):
    """An electrolyt Job.

    :ivar sandbox: The sandbox where to run the electrolyt job.
    :vartype sandbox: e3.anod.sandbox.SandBox
    :ivar force_status: Set the status of the job.
    :vartype force_status: e3.anod.status.ReturnValue
    :ivar dry_run: If True report kind of action without execution.
    :vartype dry_run: bool
    :ivar store: The store backend for accessing source and binary
        packages.
    :vartype store: e3.store.backends.base.Store
    """

    def __init__(
        self,
        uid: str,
        data: Action,
        notify_end: Callable,
        spec_repo: AnodSpecRepository,
        sandbox: SandBox,
        store: Store,
        force_status: ReturnValue = STATUS.unknown,
        dry_run: bool = False,
    ):
        """Initialize the context of the job.

        :param uid: uid of the job
        :param data: object to be processed by the job
        :param notify_end: callback to call when job is finished
        :param sandbox: Same as the attribute of the same name.
        :param store: Same as the attribute of the same name.
        :param force_status: Same as the attribute of the same name.
        :param dry_run: Same as the attribute of the same name.
        """
        super().__init__(uid, data, notify_end)
        self.__status = force_status
        self.sandbox = sandbox
        self.spec_repo = spec_repo
        self.dry_run = dry_run
        self.store = store

    def run(self) -> None:
        if self.__status == STATUS.unknown and not self.dry_run:
            try:
                getattr(self, self.data.run_method)()
            except Exception as e:
                self.__status = STATUS.failure
                logger.exception(
                    "Exception occurred in action %s %s", self.data.run_method, e
                )

    @property
    def status(self) -> ReturnValue:
        """See Job.status' description."""
        return self.__status

    def run_anod_primitive(self, primitive: str) -> None:
        """Run an anod primitive after setting up the sandbox."""
        self.data.anod_instance.sandbox = self.sandbox
        anod_driver = AnodDriver(
            anod_instance=self.data.anod_instance, store=self.store
        )
        anod_driver.activate(self.data.anod_instance.sandbox, self.spec_repo)
        anod_driver.anod_instance.build_space.create(quiet=True)
        getattr(anod_driver.anod_instance, primitive)()
        self.__status = STATUS.success

    def do_build(self) -> None:
        """Run anod build primitive."""
        return self.run_anod_primitive("build")

    def do_install(self) -> None:
        """Run anod install primitive."""
        return self.run_anod_primitive("install")

    def do_test(self) -> None:
        """Run anod test primitive."""
        return self.run_anod_primitive("test")

    def do_checkout(self) -> None:
        """Get sources from vcs to sandbox vcs_dir."""
        repo_name = self.data.repo_name
        repo_url = self.data.repo_data["url"]
        repo_revision = self.data.repo_data["revision"]
        repo_vcs = self.data.repo_data["vcs"]
        if repo_vcs != "git":
            logger.error("%s vcs type not supported", repo_vcs)
            self.__status = STATUS.failure
            return
        assert self.sandbox.vcs_dir is not None
        repo_dir = os.path.join(self.sandbox.vcs_dir, repo_name)
        g = GitRepository(repo_dir)
        if e3.log.default_output_stream is not None:
            g.log_stream = e3.log.default_output_stream
        g.init()
        g.update(repo_url, repo_revision, force=True)
        self.__status = STATUS.success

    def do_createsource(self) -> None:
        """Prepare src from vcs to cache using sourcebuilders."""
        source_name = self.data.source_name
        assert self.sandbox.tmp_dir is not None
        assert self.sandbox.vcs_dir is not None
        tmp_cache_dir = os.path.join(self.sandbox.tmp_dir, "cache")
        src = self.sandbox.vcs_dir
        src_builder = get_source_builder(
            self.data.anod_instance, source_name, local_sources_only=True
        )
        if src_builder is not None:
            repo_dict = {}
            src_dir = os.path.join(src, src_builder.checkout[0])
            dest_dir = os.path.join(tmp_cache_dir, source_name)
            # ??? missing repository state
            repo_dict[source_name] = {"working_dir": src_dir}
            mkdir(dest_dir)
            if TYPE_CHECKING:
                assert src_builder.prepare_src is not None
            src_builder.prepare_src(repo_dict, dest_dir)
            self.__status = STATUS.success
            logger.debug("%s created in cache/tmp", source_name)
        return

    def do_getsource(self) -> None:
        """action_item from an intermediate node.

        This action should return success status so do_install
        source can procede.
        """
        self.__status = STATUS.success

    def do_installsource(self) -> None:
        """Install the source from tmp/cache to build_space/src."""
        assert self.sandbox.tmp_dir is not None
        spec = self.data.spec
        spec.sandbox = self.sandbox
        anod_instance = AnodDriver(anod_instance=spec, store=self.store)
        anod_instance.activate(self.sandbox, self.spec_repo)
        source = self.data.source
        src_dir = os.path.join(self.sandbox.tmp_dir, "cache", source.name)
        if not source.dest:
            dest_dir = spec.build_space.src_dir
        else:
            dest_dir = os.path.join(spec.build_space.src_dir, source.dest)
        if not os.path.isdir(src_dir):  # defensive code
            logger.critical("source directory %s does not exist", src_dir)
            self.__status = STATUS.failure
            return
        sync_tree(src_dir, dest_dir, ignore=source.ignore)
        self.__status = STATUS.success

    def do_uploadbinarycomponent(self) -> None:
        """Upload a binary component."""
        # not implemented
        self.__status = STATUS.success

    def do_uploadsource(self) -> None:
        """Upload a binary component."""
        # not implemented
        self.__status = STATUS.success

    def do_root(self) -> None:
        """Express the final result of the exec."""
        # This method won't be executed unless all the predecessor jobs are
        # successful, so it will just report success state
        self.__status = STATUS.success
        logger.info("result: OK")


class ElectrolytJobFactory:
    def __init__(
        self,
        sandbox: SandBox,
        asr: AnodSpecRepository,
        store: Store,
        dry_run: bool = False,
    ):
        self.job_status: Dict[str, ReturnValue] = {}
        self.sandbox = sandbox
        self.asr = asr
        self.dry_run = dry_run
        self.store = store

    def get_job(
        self, uid: str, data: Action, predecessors: FrozenSet[str], notify_end: Callable
    ) -> ElectrolytJob:
        force_fail = any(
            k
            for k in predecessors
            if self.job_status[k] not in (STATUS.success, STATUS.force_skip)
        )
        return ElectrolytJob(
            uid,
            data,
            notify_end,
            spec_repo=self.asr,
            sandbox=self.sandbox,
            store=self.store,
            force_status=(STATUS.unknown if not force_fail else STATUS.failure),
            dry_run=self.dry_run,
        )

    def collect(self, job: Job) -> Literal[False]:
        """Return False as the job is never requeued."""
        self.job_status[job.uid] = job.status
        logger.info(
            "%-48s [queue=%-10s status=%-10s]",
            job.data,
            job.queue_name,
            self.job_status[job.uid].name,
        )
        return False

    def run(self, action_list: DAG) -> None:
        sch = Scheduler(self.get_job, self.collect)
        sch.run(action_list)



from e3.error import E3Error


def test_e3error():
    err = None

    try:
        raise E3Error(None)
    except E3Error as basicerr:
        assert str(basicerr) == "E3Error"

    try:
        raise E3Error(None, origin="here")
    except E3Error as err0:
        err = err0
        assert str(err).strip() == "here: E3Error"

    try:
        raise E3Error("one", origin="here")
    except E3Error as err1:
        err += err1

    try:
        raise E3Error(["two"])
    except E3Error as err2:
        err += err2
    assert str(err).strip() == "here: two"

    assert err.messages == ["one", "two"]

    err += "three"
    assert err.messages == ["one", "two", "three"]



from __future__ import annotations

import sys
from typing import overload, TYPE_CHECKING

import e3.log
import stevedore
from e3.anod.error import SandBoxError
from e3.main import Main


if TYPE_CHECKING:
    from typing import Literal, Optional
    from argparse import ArgumentParser

logger = e3.log.getLogger("sandbox.main")


@overload
def main(get_argument_parser: Literal[True]) -> ArgumentParser:
    ...


@overload
def main(get_argument_parser: Literal[False] = False) -> None:
    ...


def main(get_argument_parser: bool = False) -> Optional[ArgumentParser]:
    """Manipulate an Anod sandbox.

    This function creates the main code for the entry-point e3-sandbox. To
    create new actions it is possible to create new sandbox plugins. e.g. to
    add a new plugin ``foo`` from a package ``e3-contrib``, derives the
    class :class:`SandBoxAction` and register the extension by adding in
    :file:`e3-contrib/setup.py`::

        entry_points={
            'e3.anod.sandbox.sandbox_action': [
                'foo = e3_contrib.sandbox_actions.SandBoxFoo']
        }

    :param get_argument_parser: return e3.main.Main argument_parser instead
        of running the action.
    """
    m = Main()
    m.parse_args(known_args_only=True)

    subparsers = m.argument_parser.add_subparsers(
        title="action", description="valid actions"
    )

    # Load all sandbox actions plugins
    ext = stevedore.ExtensionManager(
        namespace="e3.anod.sandbox.sandbox_action",
        invoke_on_load=True,
        invoke_args=(subparsers,),
    )

    if len(ext.names()) != len(ext.entry_points_names()):
        raise SandBoxError(
            "an error occured when loading sandbox_action entry points %s"
            % ",".join(ext.entry_points_names())
        )  # defensive code

    if get_argument_parser:
        return m.argument_parser

    args = m.argument_parser.parse_args()

    e3.log.debug("sandbox action plugins loaded: %s", ",".join(ext.names()))

    # An action has been selected, run it
    try:
        ext[args.action].obj.run(args)
    except SandBoxError as err:
        logger.error(err)
        sys.exit(1)
    return None



from __future__ import annotations

import os
from distutils.version import StrictVersion
from typing import TYPE_CHECKING

import yaml

import e3.anod.deps
import e3.anod.package
import e3.env
import e3.log
import e3.os.process
import e3.text
from e3.anod.error import AnodError, ShellError
from e3.yaml import load_with_config

# CURRENT API version
__version__ = "1.4"

SUPPORTED_API = (__version__, "1.5")
# The driver can support multiple version of the spec API, we currently support
# only the version 1.4 and 1.5. Default is still 1.4

logger = e3.log.getLogger("anod")


if TYPE_CHECKING:
    from typing import Any, Callable, Dict, IO, List, Optional, Sequence, Union
    from e3.anod.buildspace import BuildSpace
    from e3.anod.sandbox import SandBox
    from e3.env import BaseEnv
    from e3.os.process import STDOUT_VALUE, DEVNULL_VALUE, PIPE_VALUE

    import e3.anod.package
    import e3.anod.sandbox


def check_api_version(version: str) -> None:
    """Make sure there are no API mismatch.

    :raise: AnodError if the API is not supported
    """
    if version.strip() not in SUPPORTED_API:
        raise AnodError(
            origin="check_api_version",
            message="API version mismatch. Anod specs are at %s but the "
            "driver is only supporting %s" % (version.strip(), ",".join(SUPPORTED_API)),
        )


def parse_command(command: Sequence[str], build_space: BuildSpace) -> List[str]:
    """Parse a command line formatting each string.

    :param command: the command line (a list of string)
    :param build_space: a build space object
    """
    cmd_dict = {}
    cmd_dict.update({k.upper(): v for (k, v) in list(build_space.__dict__.items())})
    return [e3.text.format_with_dict(c, cmd_dict) for c in command]


def has_primitive(anod_instance: Anod, name: str) -> bool:
    """Return True if the primitive `name` is supported.

    :param anod_instance: an Anod instance
    :param name: name of the primitive ('build', 'install'...)
    """
    if name == "source":
        source_pkg_build = getattr(anod_instance, "source_pkg_build", None)
        return source_pkg_build is not None

    try:
        func = getattr(anod_instance, name)
        is_primitive: bool = func.is_primitive
    except AttributeError:
        return False
    return is_primitive


class Anod:
    """Anod base class.

    To write an Anod specification file, you'll need to subclass Anod. A very
    basic Anod specification file could be:

    .. code-block:: python

        from e3.anod.spec import Anod

        class MyProduct(Anod):
            pass

    All attributes starting with spec_ are reserved by the driver and must not
    be overwritten.

    Several attributes are set when loading the spec:

    :cvar api_version: which API version is in use
    :cvar data_files: set of yaml files associated with the spec
    :cvar spec_checksum: the sha1 of the specification file content
    :cvar spec_dir: directory where the specification files are located
    :cvar name: the basename of the specification file (without the .anod
         extension)
    :cvar sandbox: e3.anod.sandbox.SandBox object shared by all Anod instances
    :vartype sandbox: e3.anod.sandbox.SandBox | None

    Some attributes are meant the be overwritten in the specification file:

    :cvar source_pkg_build: a dictionary associating Anod.SourceBuilder to the
        Anod.Source names

    Some attributes are here to simply the writing of specification files.
    They are part of the Anod API:

    :cvar Dependency: the e3.anod.deps.Dependency class
    :cvar Package: the e3.anod.package.Package class
    :cvar Source: the e3.anod.package.Source class
    :cvar SourceBuilder: the e3.anod.package.SourceBuilder class
    :cvar ThirdPartySourceBuilder: the e3.anod.package.ThirdPartySourceBuilder

    :ivar uid: unique identifier for the instance, None until the instance
        has been activated with AnodDriver.activate()
    :vartype uid: str | None
    """

    # set when loading the spec
    spec_checksum = ""
    spec_dir = ""
    sandbox: Optional[SandBox] = None
    name = ""
    api_version = ""
    data_files = ()

    # API
    Dependency = e3.anod.deps.Dependency
    Package = e3.anod.package.Package
    BuildVar = e3.anod.deps.BuildVar
    Source = e3.anod.package.Source
    SharedSource = e3.anod.package.SharedSource
    SourceBuilder = e3.anod.package.SourceBuilder
    ExternalSourceBuilder = e3.anod.package.ExternalSourceBuilder
    ThirdPartySourceBuilder = e3.anod.package.ThirdPartySourceBuilder

    def __init__(
        self, qualifier: str, kind: str, jobs: int = 1, env: Optional[BaseEnv] = None
    ):
        """Initialize an Anod instance.

        :param qualifier: the qualifier used when loading the spec
        :param kind: the action kind (build, install, test, ...)
        :param jobs: max parallelism level allowed for jobs spawned by this
            instance
        :param env: alternate platform environment
        :raise: SpecError
        """
        self.deps: Dict[str, Anod] = {}

        self.kind = kind
        self.jobs = jobs

        # Set when self.bind_to_sandbox is called
        self.__build_space: Optional[BuildSpace] = None

        # Default spec logger
        self.log = e3.log.getLogger("anod.spec")

        # Set spec environment
        self.env = e3.env.BaseEnv.from_env(env)

        # Create the parsed qualifier (dict version). In the future
        # self.parsed_qualifier should be replaced by self.qualifier
        self.parsed_qualifier = {}
        if qualifier:
            qual_dict = [
                (key, value)
                for key, _, value in (
                    item.partition("=") for item in qualifier.split(",")
                )
            ]
        else:
            qual_dict = []

        for k in qual_dict:
            self.parsed_qualifier[k[0]] = k[1]
        self.qualifier = qualifier

        # Default build space name is the spec name
        if "build_space_name" not in dir(self):
            self.build_space_name = self.name

        # UID of the spec instance
        self.uid = ".".join(
            (
                self.env.build.machine,
                self.env.platform,
                self.build_space_name,
                self.kind,
            )
        )

        # Hold the config dictionary-like object
        self._config: Optional[dict] = None

        # Hold the result of the pre function
        self._pre: Optional[Dict[str, Any]] = None

    @property
    def build_space(self) -> BuildSpace:
        if self.__build_space is None:
            raise AnodError("build space not set")
        return self.__build_space

    @property
    def has_package(self) -> bool:
        """Return true if the spec defines a binary package."""
        return self.package is not None and self.component is not None

    def bind_to_sandbox(self, sandbox: SandBox) -> None:
        """Bind spec instance to a physical Anod sandbox.

        Binding an Anod instance to a sandbox will set the
        build_space attribute.
        """
        self.__build_space = sandbox.get_build_space(
            name=self.build_space_name, platform=self.env.platform
        )

    def bind_to_config(self, config: dict) -> None:
        """Bind an Anod instance to a config.

        :param config: a dictionary-like object
        """
        self._config = config

    def load_config_file(
        self,
        extended: bool = False,
        suffix: Optional[str] = None,
        selectors: Optional[dict] = None,
    ) -> Any:
        """Load a YAML config file associated with the current module.

        This function looks for a YAML starting with the spec basename. The
        list of available file is set by the data_files parameter when
        initializing the spec.

        :param suffix: suffix of the configuration file (default is '')
        :param extended: if True then a special yaml parser is used with
            ability to use case statement
        :param selectors: additional selectors for extended mode
        """
        # Compute data file location and check for existence
        if StrictVersion(self.api_version) >= StrictVersion("1.5"):
            filename = os.path.join(self.name, suffix if suffix else "config")
        else:
            filename = "{}{}".format(self.name, "-" + suffix if suffix else "")
        assert filename in self.data_files, "invalid data file: {} ({})".format(
            filename, ", ".join(self.data_files),
        )
        filename = os.path.join(self.spec_dir, filename + ".yaml")

        if extended:
            # Ensure selectors is a dict
            config_selectors: dict = {} if selectors is None else selectors

            # Add environment information to selectors
            config_selectors.update(self.env.to_dict())

            return load_with_config(filename, config_selectors)
        else:
            with open(filename) as f:
                return yaml.safe_load(f.read())

    def __getitem__(self, key: str) -> Any:
        """Access build_space attributes and pre callback values directly.

        Allow accessing all build_space attributes directly by using
        __getitem__, e.g. self['PKG_DIR'] to access
        self.build_space.pkg_dir values.

        Also directly access items returned by the ``pre`` callback.
        """
        if self.__build_space is None:
            return "unknown"

        # First look for pre result
        if self._pre is not None and key in self._pre:
            return self._pre[key]

        # Then look in the config dictionary
        if self._config is not None and key in self._config:
            return self._config[key]

        # Then check if the key (in lowercase) is in the build_space
        elif key.isupper():
            return getattr(self.build_space, key.lower(), None)

    @classmethod
    def primitive(
        cls,
        pre: Optional[Callable[[Anod], dict]] = None,
        post: Optional[Callable[..., None]] = None,
        version: Optional[Callable[..., str]] = None,
    ) -> Callable:
        """Declare an anod primitive.

        Catch all exceptions and raise AnodError with the traceback

        :param pre: None or a special function to call before running the
            primitive. The function takes a unique parameter `self` and
            returns a dict
        :param post: None or a callback function to call after running the
            primitive
        :param version: None or a callback function returning the version
            that will be evaluated as a string. This callback is called
            after running the primitive
        :raise: AnodError
        """

        def primitive_dec(f, pre=pre, post=post, version=version):
            def primitive_func(self, *args, **kwargs):
                self.log.debug("%s %s starts", self.name, f.__name__)

                result = False

                # Ensure temporary directory is set to a directory local to
                # the current sandbox. This avoid mainly to loose track of
                # temporary files that are then accumulating on the
                # filesystems.
                # ??? Temporary fix for T409-012 ???
                if self.__build_space is not None and self.build_space.initialized:
                    for tmp_var in ("TMP", "TEMP", "TMPDIR"):
                        os.environ[tmp_var] = self.build_space.tmp_dir

                try:
                    # If there is a pre function call it
                    if pre is not None:
                        self._pre = getattr(self, pre)()

                    # Run the primitive
                    result = f(self, *args, **kwargs)
                    self.log.debug("%s %s ends", self.name, f.__name__)

                    # And return the result
                    return result
                except AnodError:
                    self.log.exception("%s %s fails", self.name, f.__name__)
                    raise AnodError(
                        "%s %s fails (AnodError exception in primitive)"
                        % (self.name, f.__name__)
                    )
                except Exception as e:
                    self.log.exception("%s %s fails", self.name, f.__name__)
                    raise AnodError(
                        f"{self.name} {f.__name__} fails (got exception: {e})"
                    )

            primitive_func.is_primitive = True
            primitive_func.pre = pre
            primitive_func.post = post
            primitive_func.version = version
            return primitive_func

        return primitive_dec

    @property
    def package(self) -> Optional[e3.anod.package.Package]:
        """Return binary package creation recipe.

        If None don't create a binary package, needs a component name set.
        """
        return None

    @property
    def component(self) -> Optional[str]:
        """Return component name.

        If None, don't created a component (nor a binary package).
        """
        return None

    @property
    def module_name(self) -> str:
        """For backward compatibility purpose."""
        return self.name

    @property
    def anod_id(self) -> str:
        """For backward compativility purpose."""
        return self.uid

    @property
    def source_pkg_build(self) -> Optional[List[e3.anod.package.SourceBuilder]]:
        """Return list of SourceBuilder defined in the specification file."""
        return None

    def shell(
        self,
        *command: str,
        parse_shebang: bool = True,
        output: Union[STDOUT_VALUE, DEVNULL_VALUE, PIPE_VALUE, str, IO, None] = None,
        python_executable: None = None,
        **kwargs: Any,
    ) -> e3.os.process.Run:
        """Run a subprocess using e3.os.process.Run.

        Contrary to what is done in e3.os.process.Run parse_shebang
        defaults to True and output is by default set to the anod
        build space log stream.

        Note that calling shell() raises an exception when the
        process returns an exit code that is not 0.

        Same options as e3.os.process.Run with some small differences:

        :param python_executable: kept for backward compatibility but ignored
        :param output: by default set to anod build space log stream
        :param parse_shebang: by default set True

        :raise: ShellError
        """
        parsed_command = parse_command(command, self.build_space)
        if output is None:
            output = e3.log.default_output_stream

        r = e3.os.process.Run(
            parsed_command, parse_shebang=parse_shebang, output=output, **kwargs
        )
        if TYPE_CHECKING:
            assert r.status is not None
        if r.status != 0:
            raise ShellError(
                message="%s failed (exit status: %d)" % (" ".join(command), r.status),
                origin="anod.shell",
                process=r,
            )
        return r



import e3.env
import e3.platform

import pytest


def test_platform():

    a = e3.platform.Platform.get()
    b = e3.platform.Platform.get()

    assert b == a

    assert hash(b) == hash(a)

    c = e3.platform.Platform.get(platform_name="arm-linux")

    assert b != c

    assert c.os.name == "linux"


def test_is_host():
    p = e3.platform.Platform.get(machine=e3.env.Env().build.machine)
    assert p.is_host


def test_immutable():

    a = e3.platform.Platform.get()

    with pytest.raises(AttributeError):
        a.domain = "example.net"

    b = a._replace(domain="example.net")
    assert b != a



import contextlib
import os
import sys
import time
from datetime import datetime, timedelta

from e3.fs import mkdir, rm
from e3.os.fs import touch

import pytest

if sys.platform == "win32":
    from e3.os.windows.fs import NTFile
    from e3.os.windows.native_api import (
        Access,
        FileTime,
        NTException,
        Share,
        FileAttribute,
    )

is_appveyor_test = os.environ.get("APPVEYOR") == "True"


@pytest.mark.skipif(
    sys.platform != "win32" or is_appveyor_test,
    reason="windows specific test (not working on appveyor)",
)
def test_read_attributes():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_read_attr_file.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    assert "test_read_attr_file.txt" in str(ntfile)

    ntfile.read_attributes()
    # On appveyor calling as_datetime fails because
    # ntfile.basic_info.change_time is equal to 0. Ignore this until
    # we have a real reproducer
    assert datetime.now() - ntfile.basic_info.change_time.as_datetime < timedelta(
        seconds=10
    )


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_file_info():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_read_attr_file.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    assert "test_read_attr_file.txt" in str(ntfile)

    ntfile.read_attributes()
    assert "change_time:" in str(ntfile.basic_info)


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_write_attributes():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_write_attr_file.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    ntfile.read_attributes()
    ntfile.open(Access.READ_ATTRS)
    ntfile.basic_info.change_time = FileTime(datetime.now() - timedelta(seconds=3600))
    assert str(time.localtime().tm_year) in str(ntfile.basic_info.change_time)
    try:
        with pytest.raises(NTException):
            ntfile.write_attributes()
    finally:
        ntfile.close()
    ntfile.basic_info.change_time = FileTime(datetime.now() - timedelta(days=3))
    ntfile.write_attributes()
    assert datetime.now() - ntfile.basic_info.change_time.as_datetime > timedelta(
        seconds=3000
    )


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_uid():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_uid.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    ntfile.read_attributes()
    assert ntfile.uid > 0

    ntfile2 = NTFile(os.path.join(work_dir, "non_existing.txt"))

    with pytest.raises(NTException):
        print(ntfile2.uid)


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_open_file_in_dir():
    work_dir = os.getcwd()

    test_dir_path = os.path.join(work_dir, "dir")
    mkdir(test_dir_path)
    touch(os.path.join(test_dir_path, "toto.txt"))

    with contextlib.closing(NTFile(test_dir_path)) as ntfile:
        ntfile.open()
        with contextlib.closing(NTFile("toto.txt", parent=ntfile)) as ntfile2:
            ntfile2.open()


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_volume_path():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_vpath.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    assert ntfile.volume_path

    with pytest.raises(NTException):
        # Choose a volume name that is unlikely to exist 0:/
        ntfile = NTFile("0:/dummy")
        print(ntfile.volume_path)


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_move_to_trash():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_mv_to_trash.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    ntfile.open(Access.READ_DATA)
    try:
        with pytest.raises(NTException):
            ntfile.move_to_trash()
    finally:
        ntfile.close()
    trash_path = ntfile.trash_path
    ntfile.move_to_trash()
    rm(trash_path)


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_dispose():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_dispose.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    ntfile.open(Access.READ_DATA)
    try:
        with pytest.raises(NTException):
            ntfile.dispose()
    finally:
        ntfile.close()
    ntfile.dispose()


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_rename():
    work_dir = os.getcwd()

    test_file_path = os.path.join(work_dir, "test_rename.txt")
    touch(test_file_path)
    ntfile = NTFile(test_file_path)
    ntfile.open(Access.READ_DATA)
    try:
        with pytest.raises(NTException):
            ntfile.rename(os.path.join(work_dir, "test_rename2.txt"))
    finally:
        ntfile.close()
    ntfile.rename(os.path.join(work_dir, "test_rename2.txt"))


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_iterate_on_dir():
    work_dir = os.getcwd()

    test_dir_path = os.path.join(work_dir, "dir")
    mkdir(test_dir_path)

    result = set()

    def fun(name, ntfile_instance):
        del ntfile_instance
        result.add(name)
        return True, False

    try:
        ntfile = NTFile(test_dir_path)
        status = ntfile.iterate_on_dir(fun, default_result=False)
        assert not result
        assert not status
    finally:
        ntfile.close()

    for n in range(0, 40):
        touch(os.path.join(test_dir_path, "%s.txt" % n))
    try:
        ntfile = NTFile(test_dir_path)
        status = ntfile.iterate_on_dir(fun, default_result=False)
        assert status
        assert len(result) == 40
    finally:
        ntfile.close()

    test_file = os.path.join(test_dir_path, "not-a-directory.txt")
    touch(test_file)
    ntfile = NTFile(test_file)
    try:

        with pytest.raises(NTException):
            status = ntfile.iterate_on_dir(fun, default_result=False)
    finally:
        ntfile.close()


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_is_dir_empty():
    work_dir = os.getcwd()

    test_dir_path = os.path.join(work_dir, "dir")
    deleted_file_path = os.path.join(test_dir_path, "deleted2.txt")
    deleted_file2_path = os.path.join(test_dir_path, "deleted.txt")
    mkdir(test_dir_path)

    ntfile = NTFile(test_dir_path)
    ntfile2 = NTFile(deleted_file_path)

    try:

        assert ntfile.is_dir_empty
        touch(deleted_file_path)
        touch(deleted_file2_path)
        assert not ntfile.is_dir_empty
        ntfile2.open(Access.DELETE, Share.DELETE)
        ntfile2.dispose()
        assert not ntfile.is_dir_empty
        rm(deleted_file2_path)
        assert ntfile.is_dir_empty

    finally:
        ntfile.close()
        ntfile2.close()


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_unlink():
    work_dir = os.getcwd()

    test_dir_path = os.path.join(work_dir, "dir")
    deleted_file_path = os.path.join(test_dir_path, "deleted2.txt")
    mkdir(test_dir_path)

    ntfile = NTFile(test_dir_path)
    ntfile3 = NTFile(test_dir_path)
    ntfile2 = NTFile(deleted_file_path)

    try:
        # delete inexisting file
        ntfile2.unlink()

        # delete file with readonly attribute
        touch(deleted_file_path)
        ntfile2.read_attributes()
        ntfile2.basic_info.file_attributes.attr |= FileAttribute.READONLY

        assert "READONLY" in str(ntfile2.basic_info.file_attributes)
        ntfile2.write_attributes()
        ntfile2.unlink()

        # delete file already pending deletion
        touch(deleted_file_path)
        ntfile2.open(Access.DELETE, Share.DELETE)
        ntfile2.dispose()
        ntfile2.unlink()

        # delete containing directory
        ntfile.unlink()

        ntfile.close()
        ntfile2.close()

        mkdir(test_dir_path)
        ntfile.open(Access.LIST_DIRECTORY, Share.ALL)
        ntfile3.unlink()

    finally:
        ntfile.close()
        ntfile2.close()
        ntfile2.close()

    ntfile = NTFile("nul")
    with pytest.raises(NTException) as err:
        ntfile.unlink()
    ntfile.close()
    assert "NTFile.read_attributes_internal:" in str(err.value)

    # A directory that is not empty cannot be deleted
    dir_to_delete = os.path.join(test_dir_path, "dir_to_delete")
    mkdir(dir_to_delete)
    touch(os.path.join(dir_to_delete, "afile.txt"))
    ntfile = NTFile(dir_to_delete)
    try:
        with pytest.raises(NTException) as err:
            ntfile.unlink()
    finally:
        ntfile.close()

    # A directory that is already opened and not empty cannot be
    # moved to trash
    dir_to_delete = os.path.join(test_dir_path, "dir_to_delete")
    mkdir(dir_to_delete)
    touch(os.path.join(dir_to_delete, "afile.txt"))

    ntfile = NTFile(dir_to_delete)
    ntfile2 = NTFile(dir_to_delete)
    try:
        ntfile.open(Access.LIST_DIRECTORY, Share.ALL)
        with pytest.raises(NTException) as err:
            ntfile2.unlink()
    finally:
        ntfile.close()
        ntfile2.close()

    # Try to delete a file that we cannot open
    ntfile = NTFile(deleted_file_path)
    ntfile2 = NTFile(deleted_file_path)
    try:
        touch(deleted_file_path)
        ntfile.open(Access.READ_DATA, Share.NOTHING)
        with pytest.raises(NTException) as err:
            ntfile2.unlink()
    finally:
        ntfile.close()
        ntfile2.close()



from e3.collection.dag import DAG, DAGError, DAGIterator

import pytest


def test_simple_dag():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b")
    d.add_vertex("c")
    result = []
    for vertex_id, _ in d:
        result.append(vertex_id)
    result.sort()
    assert result == ["a", "b", "c"]
    assert d.check() is None


def test_add_vertex():
    d = DAG()

    # add_vertex should fail in case a dep does not exist
    with pytest.raises(DAGError):
        d.add_vertex("a", predecessors=["b"])

    # check order of a iteration with simple dependency
    d.add_vertex("b")
    d.add_vertex("a", predecessors=["b"])

    result = []
    for vertex_id, _ in d:
        result.append(vertex_id)
    assert result == ["b", "a"]

    # check that add_vertex fails on attempt to add already existing nodde
    with pytest.raises(DAGError):
        d.add_vertex("a")

    # check update with new dependency
    d.add_vertex("c")
    d.update_vertex("b", predecessors=["c"])

    assert d.get_predecessors("b") == frozenset(["c"])
    assert d.vertex_predecessors == {
        "a": frozenset(["b"]),
        "b": frozenset(["c"]),
        "c": frozenset([]),
    }

    result = []
    for vertex_id, _ in d:
        result.append(vertex_id)
    assert result == ["c", "b", "a"]

    d.update_vertex("a", data="datafora_")
    d.update_vertex("c", data="dataforc_")
    result = []
    compound_data = ""
    for vertex_id, data in d:
        if data is not None:
            compound_data += data
        result.append(vertex_id)
    assert result == ["c", "b", "a"]
    assert compound_data == "dataforc_datafora_"


def test_cycle_detection():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b")
    d.update_vertex("a", predecessors=["b"])
    with pytest.raises(DAGError):
        d.update_vertex("b", data="newb", predecessors=["a"])

    # Ensure that DAG is still valid and that previous
    # update_vertex has no effect
    result = []
    for vertex_id, data in d:
        result.append(vertex_id)
        assert data is None
    assert result == ["b", "a"]

    # Force the creation of a cycle
    d.update_vertex("b", data="newb", predecessors=["a"], enable_checks=False)

    # Verify that the cycle is detected
    with pytest.raises(DAGError):
        d.check()

    # Verify that some functions do not hang when a cycle is present
    assert len(d.get_closure("b")) == 2
    assert str(d)
    assert d.as_dot()

    with pytest.raises(DAGError):
        d.reverse_graph()


def test_shortest_path():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b")
    d.update_vertex("a", predecessors=["b"])
    d.update_vertex("b", predecessors=["a"], enable_checks=False)
    assert d.shortest_path("a", "a") == ["a", "b", "a"]

    d = DAG()
    d.add_vertex("a")
    d.update_vertex("a", predecessors=["a"], enable_checks=False)
    assert d.shortest_path("a", "a") == ["a", "a"]

    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b", predecessors=["a"])
    d.add_vertex("c", predecessors=["b"])
    d.add_vertex("d", predecessors=["c", "a"])
    assert d.shortest_path("a", "d") == ["a", "d"]
    assert d.shortest_path("d", "a") is None


def test_dag_merge():
    d = DAG()
    d.add_vertex("b")
    d.add_vertex("a", predecessors=["b"])

    d2 = DAG()
    d2.add_vertex("c")
    d2.add_vertex("b", predecessors=["c"])
    d2.add_vertex("a", predecessors=["c"])

    d3 = d | d2
    result = []
    for vertex_id, data in d3:
        result.append(vertex_id)
        assert data is None
    assert result == ["c", "b", "a"]


def test_dag_len():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b")
    d.update_vertex("a", predecessors=["b"])
    assert len(d) == 2


def test_dag_str():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b")
    d.update_vertex("a", predecessors=["b"])
    assert str(d)


def test_iter_with_busy_state():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b", predecessors=["a"])

    it = DAGIterator(d, enable_busy_state=True)
    for nid, _ in it:
        if nid is None:
            it.leave("a")


def test_inexisting():
    d = DAG()
    d.add_vertex("a")
    assert "a" in d
    d.update_vertex("a", data="NOT B", predecessors=["b"], enable_checks=False)
    assert "b" not in d
    assert d["a"] == "NOT B"
    with pytest.raises(DAGError):
        d.check()


def test_cycle():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b")
    d.update_vertex("a", predecessors=["b"])
    d.add_vertex("c", predecessors=["b"])
    d.update_vertex("b", predecessors=["c"], enable_checks=False)
    with pytest.raises(DAGError):
        d.check()

    with pytest.raises(DAGError):
        d.get_context("b")


def test_reverse_dag():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b", predecessors=["a"])
    d.add_vertex("c", predecessors=["b"])
    d.add_vertex("d", predecessors=["c"])

    it = DAGIterator(d)
    assert [k for k, _ in it] == ["a", "b", "c", "d"]

    reverse_d = d.reverse_graph()
    reverse_it = DAGIterator(reverse_d)
    assert [k for k, _ in reverse_it] == ["d", "c", "b", "a"]


def test_dot():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b", predecessors=["a"])
    assert '"b" -> "a"' in d.as_dot()


def test_pruned_dag():
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("a1")
    d.add_vertex("a2")
    d.add_vertex("b", predecessors=["a", "a1"])
    d.add_vertex("c", predecessors=["b", "a2"])
    d.add_vertex("d", predecessors=["c"])
    d.add_tag("d", "tag")

    def f(dg, node):
        if node in ("b", "c"):
            return True
        return False

    d2 = d.prune(f)

    for node, _ in d2:
        preds = d2.get_predecessors(node)
        if node == "d":
            assert len(preds) == 3
            assert "a" in preds
            assert "a1" in preds
            assert "a2" in preds
            assert d2.get_tag("d") == "tag"
        elif node.startswith("a"):
            assert len(preds) == 0
        else:
            raise AssertionError("invalid node")

    # By default prune ensure we keep context unchanged.
    # If we add a tag to 'b' (that would be suppressed) then we expect a
    # DAGError
    d.add_tag("b", "b_tag")
    with pytest.raises(DAGError):
        d.prune(f)

    # In that last test, check that no error is raised but that tag are
    # preserved
    d4 = d.prune(f, preserve_context=False)
    assert d4.get_tag("d") == "tag"


def test_tagged_dag():
    r"""Test add_tag/get_tag/get_context.

    With the following DAG::

               A
              / \
             B   C*
           /  \ /
          D*   E
         / \  / \
        F   G    H*
    """
    d = DAG()
    d.add_vertex("a")
    d.add_vertex("b", predecessors=["a"])
    d.add_vertex("c", predecessors=["a"])
    d.add_vertex("d", predecessors=["b"])
    d.add_vertex("e", predecessors=["b", "c"])
    d.add_vertex("f", predecessors=["d"])
    d.add_vertex("g", predecessors=["d", "e"])
    d.add_vertex("h", predecessors=["e"])

    d.add_tag("c", data="tagc")
    d.add_tag("d", data="tagd")
    d.add_tag("h", data="tagh")

    assert d.get_tag("a") is None
    assert d.get_tag("b") is None
    assert d.get_tag("c") == "tagc"
    assert d.get_tag("e") is None
    assert d.get_tag("h") == "tagh"

    assert d.get_context("d") == [(0, "d", "tagd")]
    assert d.get_context("g") == [(1, "d", "tagd"), (2, "c", "tagc")]
    assert d.get_context("f") == [(1, "d", "tagd")]
    assert d.get_context("b") == []
    assert d.get_context("a") == []
    assert d.get_context("c") == [(0, "c", "tagc")]
    assert d.get_context("e") == [(1, "c", "tagc")]
    assert d.get_context("h") == [(0, "h", "tagh")]

    assert d.get_context("e", reverse_order=True) == [(1, "h", "tagh")]
    assert d.get_context("h", reverse_order=True) == [(0, "h", "tagh")]
    assert d.get_context("a", reverse_order=True) == [
        (1, "c", "tagc"),
        (2, "d", "tagd"),
        (3, "h", "tagh"),
    ]

    assert d.get_context("a", reverse_order=True) == [
        (1, "c", "tagc"),
        (2, "d", "tagd"),
        (3, "h", "tagh"),
    ]

    assert d.get_context(vertex_id="a", max_distance=2, reverse_order=True) == [
        (1, "c", "tagc"),
        (2, "d", "tagd"),
    ]

    assert d.get_context(vertex_id="a", max_element=2, reverse_order=True) == [
        (1, "c", "tagc"),
        (2, "d", "tagd"),
    ]



import sys

import e3.os.process

import pytest

if sys.platform == "win32":
    from e3.os.windows.process import wait_for_objects, process_exit_code


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_invalid_handle():
    with pytest.raises(WindowsError):
        process_exit_code(42)


@pytest.mark.skipif(sys.platform != "win32", reason="windows specific test")
def test_wait_for_objects():
    long_cmd = [sys.executable, "-c", "import time; time.sleep(40.0)"]
    short_cmd = [sys.executable, "-c", "pass"]

    p = e3.os.process.Run(long_cmd, bg=True)

    assert (
        wait_for_objects([int(p.internal._handle)], timeout=1) is None
    ), "timeout was expected"
    p.kill()

    p0 = e3.os.process.Run(long_cmd, bg=True)
    p1 = e3.os.process.Run(short_cmd, bg=True)

    try:
        assert (
            wait_for_objects(
                [int(p0.internal._handle), int(p1.internal._handle)], timeout=2
            )
            == 1
        ), "process 1 was expected"
    finally:
        p0.kill()
        p1.kill()

    p0 = e3.os.process.Run(long_cmd, bg=True)
    p1 = e3.os.process.Run(short_cmd, bg=True)

    try:
        assert (
            wait_for_objects(
                [int(p0.internal._handle), int(p1.internal._handle)],
                timeout=2,
                wait_for_all=True,
            )
            is None
        ), "timeout expected"
    finally:
        p0.kill()
        p1.kill()

    p0 = e3.os.process.Run(short_cmd, bg=True)
    p1 = e3.os.process.Run(short_cmd, bg=True)
    try:
        assert (
            wait_for_objects(
                [int(p0.internal._handle), int(p1.internal._handle)],
                timeout=0,
                wait_for_all=True,
            )
            is not None
        ), "no timeout expected"
    finally:
        p0.kill()
        p1.kill()



"""Implementation of Direct Acyclic Graphs."""


from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from e3.error import E3Error

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Dict,
        FrozenSet,
        Hashable,
        List,
        Iterator,
        Optional,
        Sequence,
        Set,
        Tuple,
        Union,
    )

    VertexID = Hashable


class DAGError(E3Error):
    pass


class DAGIterator:

    NOT_VISITED = 0
    BUSY = 1
    VISITED = 2

    def __init__(self, dag: DAG, enable_busy_state: bool = False):
        """Initialize DAG iterator.

        :param dag: the dag on which iteration is done
        :param enable_busy_state: used to implement topological parallel
            iteration. When False, a vertex has only 2 possible states: VISITED
            or NOT_VISITED. When true, calling the next() function on the
            iteratior puts the element in 'BUSY' state. It will only be moved
            to 'Visited' once calling the 'leave()' method.
        """
        self.dag = dag
        self.non_visited = set(self.dag.vertex_data.keys())
        self.states = {k: self.NOT_VISITED for k in list(self.dag.vertex_data.keys())}
        self.enable_busy_state = enable_busy_state

        # Compute number of non visited predecessors for each node.
        # Doing this computation in advance enable faster
        # iteration overall (simplify conditions in next_element)
        self.pred_number = {k: len(v) for k, v in self.dag.vertex_predecessors_items()}

    def __iter__(self) -> DAGIterator:
        return self

    def __next__(self) -> Union[Tuple[None, None], Tuple[VertexID, Any]]:
        """Retrieve next_element with with_predecessors=False.

        The intermediate function is needed in Python 3.x
        """
        vertex_id, data, _ = self.next_element()
        if vertex_id is None:
            return (None, None)
        return (vertex_id, data)

    def next_element(self,) -> Tuple[Union[None, VertexID], Any, FrozenSet[VertexID]]:
        """Retrieve next element in topological order.

        :return: a vertex id, data, predecessors. (None, None, frozenset()) is
            returned if no element is available).
        """
        if not self.non_visited:
            raise StopIteration

        # Retrieve the first vertex for which all the predecessors have been
        # visited
        result = next((k for k in self.non_visited if self.pred_number[k] == 0), None)

        if result is None:
            if not self.enable_busy_state:
                for node in self.non_visited:
                    minimal_cycle = self.dag.shortest_path(node, node)
                    if minimal_cycle is not None:
                        raise DAGError(
                            "cycle detected: %s"
                            % " -> ".join(
                                [str(vertex_id) for vertex_id in minimal_cycle]
                            )
                        )
                raise DAGError("cycle detected (unknown error)")

            # No vertex is ready to be visited
            return None, None, frozenset()

        # Remove the vertex from the "non_visited_list" and when
        # enable_busy_state, mark the vertex as BUSY, mark it VISITED
        # otherwise.
        if self.enable_busy_state:
            self.states[result] = self.BUSY
        else:
            self.states[result] = self.VISITED
            # Update the number of non visited predecessors
            for k in self.dag.get_successors(result):
                self.pred_number[k] -= 1

        self.non_visited.discard(result)

        return (result, self.dag.vertex_data[result], self.dag.get_predecessors(result))

    def leave(self, vertex_id: VertexID) -> None:
        """Switch element from BUSY to VISITED state.

        :param vertex_id: the vertex to leave
        """
        assert self.states[vertex_id] == self.BUSY
        self.states[vertex_id] = self.VISITED
        # Update the number of non visited predecessors
        for k in self.dag.get_successors(vertex_id):
            self.pred_number[k] -= 1


class DAG:
    """Represent a Directed Acyclic Graph.

    :ivar vertex_data: a dictionary containing all vertex data
        indexed by vertex id
    :vartype vertex_data: dict
    :ivar tags: a dictionary containing "tags" associated with
        a vertex data, indexed by vertex id
    """

    def __init__(self) -> None:
        """Initialize a DAG."""
        self.vertex_data: Dict[VertexID, Any] = {}
        self.tags: Dict[VertexID, Any] = {}

        self.__vertex_predecessors: Dict[VertexID, FrozenSet[VertexID]] = {}
        self.__vertex_successors: Dict[VertexID, FrozenSet[VertexID]] = {}
        self.__has_cycle: Optional[bool] = None

    @property
    def vertex_predecessors(self) -> Dict[VertexID, FrozenSet[VertexID]]:
        """Return predecessors.

        Meant only for backward compatibility. Use vertex_predecessors_items.

        :return: a dictionary containing the list of predecessors for each
            vertex, indexed by vertex id
        """
        # We're doing a copy of the __vertex_predecessors dictionary
        # to avoid external modifications
        return dict(self.__vertex_predecessors)

    def vertex_predecessors_items(
        self,
    ) -> Iterator[Tuple[VertexID, FrozenSet[VertexID]]]:
        """Return predecessors.

        :return: a list of (vertex id, predecessors)
        """
        return iter(self.__vertex_predecessors.items())

    def get_predecessors(self, vertex_id: VertexID) -> FrozenSet[VertexID]:
        """Get set of predecessors for a given vertex."""
        return self.__vertex_predecessors.get(vertex_id, frozenset())

    def set_predecessors(
        self, vertex_id: VertexID, predecessors: FrozenSet[VertexID]
    ) -> None:
        """Set predecessors for a given vertex.

        Invalidate the global dictionary of vertex successors.
        """
        self.__vertex_predecessors[vertex_id] = predecessors
        # Reset successors and cycle check results which are now invalid
        self.__vertex_successors = {}
        self.__has_cycle = None

    def get_successors(self, vertex_id: VertexID) -> FrozenSet[VertexID]:
        """Get set of successors for a given vertex.

        If the global dictionary of vertex successors has not been
        computed or if it has been invalidated then recompute it.
        """
        if self.__vertex_successors == {}:
            successors: Dict[VertexID, Set[VertexID]] = {
                k: set() for k in self.__vertex_predecessors
            }
            for pred_k, pred_v in self.__vertex_predecessors.items():
                for el in pred_v:
                    successors[el].add(pred_k)
            # Use frozenset to prevent the modification of successors
            for succ_k, succ_v in successors.items():
                self.__vertex_successors[succ_k] = frozenset(succ_v)

        return self.__vertex_successors.get(vertex_id, frozenset())

    def add_tag(self, vertex_id: VertexID, data: Any) -> None:
        """Tag a vertex.

        :param vertex_id: ID of the vertex to tag
        :param data: tag content
        """
        self.tags[vertex_id] = data

    def get_tag(self, vertex_id: VertexID) -> Any:
        """Retrieve a tag associated with a vertex.

        :param vertex_id: ID of the vertex
        :return: tag content
        """
        return self.tags.get(vertex_id)

    def get_context(
        self,
        vertex_id: VertexID,
        max_distance: Optional[int] = None,
        max_element: Optional[int] = None,
        reverse_order: bool = False,
    ) -> List[Tuple[int, VertexID, Any]]:
        r"""Get tag context.

        Returns the list of predecessors tags along with their vertex id and
        the distance between the given vertex and the tag. On each predecessors
        branch the first tag in returned. So for the following graph::


                A*
               / \
              B   C*
             / \   \
            D   E*  F

        where each node with a * are tagged

        get_context(D) will return (2, A, <tag A>)
        get_context(E) will return (0, E, <tag E>)
        get_context(F) will return (1, C, <tag C>)

        When using reverse_order=True, get_context will follow successors
        instead of predecessors.

        get_context(B, reverse_order=True) will return (1, E, <tag E>)

        :param vertex_id: ID of the vertex
        :param max_distance: do not return resultsh having a distance higher
            than ``max_distance``
        :param max_element: return only up-to ``max_element`` elements
        :param reverse_order: when True follow successors instead of
            predecessors
        :return: a list of tuple (distance:int, tagged vertex id, tag content)
        """
        self.check()

        def get_next(vid: VertexID) -> FrozenSet[VertexID]:
            """Get successors or predecessors.

            :param vid: vertex id
            """
            if reverse_order:
                result = self.get_successors(vid)
            else:
                result = self.get_predecessors(vid)
            return result

        visited: Set[VertexID] = set()
        tags = []
        distance = 0
        node_tag = self.get_tag(vertex_id)
        if node_tag is not None:
            tags.append((distance, vertex_id, node_tag))
            return tags

        closure = get_next(vertex_id)
        closure_len = len(closure)

        while True:
            distance += 1
            if max_distance is not None and distance > max_distance:
                return tags
            for n in closure - visited:
                visited.add(n)

                n_tag = self.get_tag(n)
                if n_tag is not None:
                    tags.append((distance, n, n_tag))

                    if max_element is not None and len(tags) == max_element:
                        return tags
                else:
                    # Search tag in vertex predecessors
                    closure |= get_next(n)

            if len(closure) == closure_len:
                break
            closure_len = len(closure)
        return tags

    def add_vertex(
        self,
        vertex_id: VertexID,
        data: Any = None,
        predecessors: Optional[Sequence[VertexID]] = None,
    ) -> None:
        """Add a new vertex into the DAG.

        :param vertex_id: the name of the vertex
        :param data: data for the vertex.
        :param predecessors: list of predecessors (vertex ids) or None
        :raise: DAGError if cycle is detected or else vertex already exist
        """
        if vertex_id in self.vertex_data:
            raise DAGError(
                message="vertex %s already exist" % vertex_id, origin="DAG.add_vertex"
            )
        self.update_vertex(vertex_id, data, predecessors)

    def update_vertex(
        self,
        vertex_id: VertexID,
        data: Any = None,
        predecessors: Optional[
            Union[Sequence[VertexID], Set[VertexID], FrozenSet[VertexID]]
        ] = None,
        enable_checks: bool = True,
    ) -> None:
        """Update a vertex into the DAG.

        :param vertex_id: the name of the vertex
        :param data: data for the vertex. If None and vertex already exist
            then data value is preserved
        :param predecessors: list of predecessors (vertex ids) or None. If
            vertex already exists predecessors are added to the original
            predecessors
        :param enable_checks: if False check that all predecessors exists and
            that no cycle is introduce is not perform (for performance)
        :raise: DAGError if cycle is detected
        """
        if predecessors is None:
            vertex_predecessors: FrozenSet[VertexID] = frozenset()
        else:
            vertex_predecessors = frozenset(predecessors)

        if enable_checks:
            # Before doing changes ensure that the DAG is already valid
            self.check()

            non_existing_predecessors = [
                k for k in vertex_predecessors if k not in self.vertex_data
            ]
            if non_existing_predecessors:
                raise DAGError(
                    message="predecessor on non existing vertices {}".format(
                        ", ".join([str(nid) for nid in non_existing_predecessors])
                    ),
                    origin="DAG.update_vertex",
                )

        if vertex_id not in self.vertex_data:
            self.set_predecessors(vertex_id, vertex_predecessors)
            self.vertex_data[vertex_id] = data
        else:
            previous_predecessors = self.get_predecessors(vertex_id)
            self.set_predecessors(
                vertex_id, previous_predecessors | vertex_predecessors
            )

            if enable_checks:
                # Will raise DAGError if a cycle is created
                if vertex_id in self.get_closure(vertex_id):
                    minimal_cycle = self.shortest_path(vertex_id, vertex_id)
                    if TYPE_CHECKING:
                        assert minimal_cycle is not None
                    self.set_predecessors(vertex_id, previous_predecessors)
                    raise DAGError(
                        message="cannot update vertex ({} create a cycle: {})".format(
                            vertex_id, " -> ".join([str(vid) for vid in minimal_cycle])
                        ),
                        origin="DAG.update_vertex",
                    )

            if data is not None:
                self.vertex_data[vertex_id] = data

        if not enable_checks:
            # DAG modified without cycle checks, discard cached result
            self.__has_cycle = None

    def shortest_path(
        self, source: VertexID, target: VertexID
    ) -> Optional[List[VertexID]]:
        """Compute the shortest path between two vertices of the DAG.

        :param source: vertex id of the source
        :param target: vertex id of the target. If target is equal to source
            then the algorithm try to find the shortest cycle
        :return: a list of vertex. If there is no path between two nodes then
            return None
        """
        # We use the Dikjstra algorithm to compute the shortest path.
        # Note that this is a slight variation so that the algorithm
        # can be used to compute shortest cycle on a given node.

        # The maximum distance between two vertices is len(DAG) - 1. Thus
        # len(DAG) can be our infinite distance. As when target and source
        # are equals we might add an additional fake node, increase it by one.
        infinite = len(self.vertex_data) + 1

        # Keep track of minimal distance between vertices and the sources
        dist: Dict[Optional[VertexID], int] = {k: infinite for k in self.vertex_data}

        # Keep track of the minimum distance
        prev: Dict[Optional[VertexID], Optional[VertexID]] = {
            k: None for k in self.vertex_data
        }

        # Set of non visited vertices
        unvisited: Set[Optional[VertexID]] = set(self.vertex_data)

        # The only known distance at startup
        dist[target] = 0

        path_source: Optional[VertexID] = source

        if path_source == target:
            # If source is equal to target, default dikjstra algorithm does
            # not work. Add a fake node and use it as target. When iterating
            # on predecessors replace all occurences of sources to that node.
            # If we find a path between that node and the source, it means
            # we have our shortest cycle.
            dist[None] = infinite
            prev[None] = None
            unvisited.add(None)
            path_source = None

        while unvisited:
            u = min(unvisited, key=lambda x: dist[x])
            unvisited.remove(u)

            # We have found our shortest path, so break
            if u == path_source:
                break
            elif u is None:
                continue

            for u_pred in self.get_predecessors(u):
                # Handle cycle detection case.
                if path_source is None and u_pred == target:
                    v = None
                else:
                    v = u_pred

                alt = dist[u] + 1
                if alt < dist[v]:
                    dist[v] = alt
                    prev[v] = u

        if dist[path_source] == infinite:
            # No path exist between source and target (or no cycle).
            return None
        else:
            result: List[Optional[VertexID]] = [path_source]
            while prev[result[-1]] is not None:
                result.append(prev[result[-1]])

            # In case of cycle detection the first node is None which
            # is in reality our source node.
            if path_source is None:
                result[0] = target
            return result  # type: ignore

    def check(self) -> None:
        """Check for cycles and inexisting nodes.

        :raise: DAGError if the DAG is not valid
        """
        # Noop if check already done
        if self.__has_cycle is False:
            return
        elif self.__has_cycle:
            raise DAGError(
                message="this DAG contains at least one cycle", origin="DAG.check"
            )
        # First check predecessors validity
        for node, preds in self.__vertex_predecessors.items():
            if len([k for k in preds if k not in self.vertex_data]) > 0:
                self.__has_cycle = True
                raise DAGError(
                    message="invalid nodes in predecessors of %s" % node,
                    origin="DAG.check",
                )
        # raise DAGError if cycle
        try:
            for _ in DAGIterator(self):
                pass
        except DAGError:
            self.__has_cycle = True
            raise
        else:
            self.__has_cycle = False

    def get_closure(self, vertex_id: VertexID) -> FrozenSet[VertexID]:
        """Retrieve closure of predecessors for a vertex.

        :param vertex_id: the vertex to inspect
        :return: a set of vertex_id
        """
        visited: Set[VertexID] = set()
        closure = self.get_predecessors(vertex_id)
        closure_len = len(closure)

        while True:
            for n in closure - visited:
                visited.add(n)

                closure |= self.get_predecessors(n)

            if len(closure) == closure_len:
                break
            closure_len = len(closure)
        return closure

    def reverse_graph(self) -> DAG:
        """Compute the reverse DAG.

        :return: the reverse DAG (edge inverted)
        """
        result = DAG()

        # Copy the tags to the reverse DAG
        result.tags = self.tags

        # Note that we don't need to enable checks during this operation
        # as the reverse graph of a DAG is still a DAG (no cycles).
        for node, predecessors in self.__vertex_predecessors.items():
            result.update_vertex(node, data=self.vertex_data[node], enable_checks=False)
            for p in predecessors:
                result.update_vertex(p, predecessors=[node], enable_checks=False)
        try:
            result.check()
        except DAGError:
            # Check detected
            self.__has_cycle = True
            raise
        else:
            # No cycle in the DAG
            self.__has_cycle = False
        return result

    def __iter__(self) -> DAGIterator:
        return DAGIterator(self)

    def __contains__(self, vertex_id: VertexID) -> bool:
        """Check if a vertex is present in the DAG."""
        return vertex_id in self.vertex_data

    def __getitem__(self, vertex_id: VertexID) -> Any:
        """Get data associated with a vertex."""
        return self.vertex_data[vertex_id]

    def __or__(self, other: DAG) -> DAG:
        """Merge two dags."""
        assert isinstance(other, DAG)

        result = DAG()

        # First add vertices and then update predecessors. The two step
        # procedure is needed because predecessors should exist.
        for nid in self.vertex_data:
            result.add_vertex(nid)

        for nid in other.vertex_data:
            result.update_vertex(nid)

        # Update predecessors
        for nid in self.vertex_data:
            result.update_vertex(nid, self.vertex_data[nid], self.get_predecessors(nid))

        for nid in other.vertex_data:
            result.update_vertex(
                nid, other.vertex_data[nid], other.get_predecessors(nid)
            )

        # Make sure that no cycle are created the merged DAG.
        result.check()
        return result

    def as_dot(self) -> str:
        """Return a Graphviz graph representation of the graph.

        :return: the dot source file
        """
        result = ["digraph G {", 'rankdir="LR";']
        for vertex in self.vertex_data:
            result.append('"%s"' % vertex)
        for vertex, predecessors in self.__vertex_predecessors.items():
            for predecessor in predecessors:
                result.append(f'"{vertex}" -> "{predecessor}"')
        result.append("}")
        return "\n".join(result)

    def prune(
        self, fun: Callable[[DAG, VertexID], bool], preserve_context: bool = True
    ) -> DAG:
        """Create a pruned graph.

        :param fun: function that return True whenever a node should be
            pruned. The function receive as parameter the dag and the node id
        :param preserve_context: if True ensure that context is preserved
            (i.e: that calls to get_context will return the same value for
             both current graph and pruned graph). This means that any attempt
            to remove a node containing a tag will result in a DAGError.
        :return: a new DAG
        """
        result = DAG()

        # Used to maintain the new list of predecessors
        pruned_node_predecessors: Dict[VertexID, Set[VertexID]] = {}

        for node, data in self:
            # The new list of predecessors is the union of predecessors of
            # pruned predecessors and the other predecessors.
            if TYPE_CHECKING:
                assert node is not None
            predecessors = set(
                chain(
                    *[pruned_node_predecessors[k] for k in self.get_predecessors(node)]
                )
            )

            if fun(self, node):
                # Check if node can pruned
                if node in self.tags and preserve_context:
                    raise DAGError("suppressing %s impact context" % node)

                # Node is pruned keep track of its predecessors
                pruned_node_predecessors[node] = predecessors
            else:
                # Node is kept. Note that no check is needed as pruning
                # operation cannot introduce a cycle.
                result.update_vertex(node, data, predecessors, enable_checks=False)
                pruned_node_predecessors[node] = {node}
                if node in self.tags:
                    result.add_tag(node, self.tags[node])
        return result

    def __len__(self) -> int:
        return len(self.vertex_data)

    def __str__(self) -> str:
        result = []
        for vertex, predecessors in self.__vertex_predecessors.items():
            if predecessors:
                result.append(
                    "{} -> {}".format(vertex, ", ".join(str(p) for p in predecessors))
                )
            else:
                result.append("%s -> (none)" % vertex)
        return "\n".join(result)
